from .annotation import Bitmask, BoundingBox, Detection, Polygon
from .computation import (
    compute_bbox_iou,
    compute_bitmask_iou,
    compute_confusion_matrix,
    compute_polygon_iou,
    compute_precion_recall,
    compute_ranked_pairs,
)
from .manager import DataLoader, Evaluator
from .metric import Metric, MetricType

__all__ = [
    "Bitmask",
    "BoundingBox",
    "Detection",
    "Polygon",
    "Metric",
    "MetricType",
    "compute_bbox_iou",
    "compute_bitmask_iou",
    "compute_polygon_iou",
    "compute_ranked_pairs",
    "compute_precion_recall",
    "compute_confusion_matrix",
    "DataLoader",
    "Evaluator",
]



import pytest
from valor_lite.text_generation import Context, Evaluator, QueryResponse
from valor_lite.text_generation.computation import calculate_rouge_scores


def test_rouge_no_context(mock_client):

    evaluator = Evaluator(client=mock_client)
    with pytest.raises(ValueError):
        assert evaluator.compute_rouge(
            response=QueryResponse(
                query="a",
                response="a",
            )
        )


def test_calculate_rouge_scores():

    rouge_types = [
        "rouge1",
        "rouge2",
        "rougeL",
        "rougeLsum",
    ]

    # perfect match
    assert calculate_rouge_scores(
        prediction="Mary loves Joe",
        references=[
            "Mary loves Joe",
        ],
        rouge_types=rouge_types,
        use_stemmer=False,
    ) == {
        "rouge1": 1.0,
        "rouge2": 1.0,
        "rougeL": 1.0,
        "rougeLsum": 1.0,
    }

    # perfect match, case sensitive
    assert calculate_rouge_scores(
        prediction="MARY LOVES JOE",
        references=[
            "Mary loves Joe",
        ],
        rouge_types=rouge_types,
        use_stemmer=False,
    ) == {
        "rouge1": 1.0,
        "rouge2": 1.0,
        "rougeL": 1.0,
        "rougeLsum": 1.0,
    }

    # perfect match, case sensitive
    assert calculate_rouge_scores(
        prediction="Mary loves Joe",
        references=[
            "MARY LOVES JOE",
        ],
        rouge_types=rouge_types,
        use_stemmer=False,
    ) == {
        "rouge1": 1.0,
        "rouge2": 1.0,
        "rougeL": 1.0,
        "rougeLsum": 1.0,
    }

    # off by one
    assert calculate_rouge_scores(
        prediction="Mary loves Joe",
        references=["Mary loves Jane"],
        rouge_types=rouge_types,
        use_stemmer=False,
    ) == {
        "rouge1": 2 / 3,
        "rouge2": 0.5,
        "rougeL": 2 / 3,
        "rougeLsum": 2 / 3,
    }

    # incorrect match without stemming
    assert calculate_rouge_scores(
        prediction="flipping the roaring white dolphin",
        references=["flip the roaring white dolphin"],
        rouge_types=rouge_types,
        use_stemmer=False,
    ) == {
        "rouge1": 0.8000000000000002,
        "rouge2": 0.75,
        "rougeL": 0.8000000000000002,
        "rougeLsum": 0.8000000000000002,
    }

    # correct match with stemming
    assert calculate_rouge_scores(
        prediction="flipping the roaring white dolphin",
        references=["flip the roaring white dolphin"],
        rouge_types=rouge_types,
        use_stemmer=True,
    ) == {
        "rouge1": 1,
        "rouge2": 1,
        "rougeL": 1,
        "rougeLsum": 1,
    }

    # test multiple references
    assert calculate_rouge_scores(
        prediction="flipping the roaring white dolphin",
        references=[
            "some random sentence",
            "some other sentence",
            "some final reference",
            "flip the roaring white dolphin",
        ],
        rouge_types=rouge_types,
        use_stemmer=True,
    ) == {
        "rouge1": 1,
        "rouge2": 1,
        "rougeL": 1,
        "rougeLsum": 1,
    }

    # references isn't a list
    assert calculate_rouge_scores(
        prediction="Mary loves Joe",
        references="Mary loves Joe",
        rouge_types=rouge_types,
    ) == {
        "rouge1": 1,
        "rouge2": 1,
        "rougeL": 1,
        "rougeLsum": 1,
    }

    # predictions as a list
    with pytest.raises(ValueError):
        calculate_rouge_scores(
            prediction=["Mary loves Joe"],  # type: ignore - testing
            references=["Mary loves June"],
            rouge_types=rouge_types,
        )


def test_evaluate_rouge():

    evaluator = Evaluator()

    # perfect match
    metrics = evaluator.compute_rouge(
        response=QueryResponse(
            query="n/a",
            response="Mary loves Joe",
            context=Context(
                groundtruth=["Mary loves Joe"],
            ),
        )
    )
    assert [m.to_dict() for m in metrics] == [
        {
            "type": "ROUGE",
            "value": 1.0,
            "parameters": {
                "rouge_type": "rouge1",
                "use_stemmer": False,
            },
        },
        {
            "type": "ROUGE",
            "value": 1.0,
            "parameters": {
                "rouge_type": "rouge2",
                "use_stemmer": False,
            },
        },
        {
            "type": "ROUGE",
            "value": 1.0,
            "parameters": {
                "rouge_type": "rougeL",
                "use_stemmer": False,
            },
        },
        {
            "type": "ROUGE",
            "value": 1.0,
            "parameters": {
                "rouge_type": "rougeLsum",
                "use_stemmer": False,
            },
        },
    ]

    # off by one
    metrics = evaluator.compute_rouge(
        response=QueryResponse(
            query="n/a",
            response="Mary loves Joe",
            context=Context(
                groundtruth=["Mary loves Jane"],
            ),
        )
    )
    assert [m.to_dict() for m in metrics] == [
        {
            "type": "ROUGE",
            "value": 2 / 3,
            "parameters": {
                "rouge_type": "rouge1",
                "use_stemmer": False,
            },
        },
        {
            "type": "ROUGE",
            "value": 0.5,
            "parameters": {
                "rouge_type": "rouge2",
                "use_stemmer": False,
            },
        },
        {
            "type": "ROUGE",
            "value": 2 / 3,
            "parameters": {
                "rouge_type": "rougeL",
                "use_stemmer": False,
            },
        },
        {
            "type": "ROUGE",
            "value": 2 / 3,
            "parameters": {
                "rouge_type": "rougeLsum",
                "use_stemmer": False,
            },
        },
    ]



from valor_lite.semantic_segmentation import (
    DataLoader,
    Metric,
    MetricType,
    Segmentation,
)


def test_accuracy_basic_segmentations(basic_segmentations: list[Segmentation]):
    loader = DataLoader()
    loader.add_data(basic_segmentations)
    evaluator = loader.finalize()

    metrics = evaluator.evaluate()

    actual_metrics = [m.to_dict() for m in metrics[MetricType.Accuracy]]
    expected_metrics = [
        {
            "type": "Accuracy",
            "value": 0.5,
            "parameters": {},
        },
    ]
    for m in actual_metrics:
        assert m in expected_metrics
    for m in expected_metrics:
        assert m in actual_metrics


def test_accuracy_segmentations_from_boxes(
    segmentations_from_boxes: list[Segmentation],
):
    loader = DataLoader()
    loader.add_data(segmentations_from_boxes)
    evaluator = loader.finalize()

    metrics = evaluator.evaluate()

    actual_metrics = [m.to_dict() for m in metrics[MetricType.Accuracy]]
    expected_metrics = [
        {
            "type": "Accuracy",
            "value": 0.9444481481481481,
            "parameters": {},
        },
    ]
    for m in actual_metrics:
        assert m in expected_metrics
    for m in expected_metrics:
        assert m in actual_metrics


def test_accuracy_large_random_segmentations(
    large_random_segmentations: list[Segmentation],
):
    loader = DataLoader()
    loader.add_data(large_random_segmentations)
    evaluator = loader.finalize()

    metrics = evaluator.evaluate()[MetricType.Accuracy]

    assert len(metrics) == 1
    assert isinstance(metrics[0], Metric)
    assert (
        round(metrics[0].value, 1)  # type: ignore - testing
        == 0.5  # random choice
    )



from dataclasses import dataclass


@dataclass
class Classification:
    """
    Classification data structure containing a ground truth label and a list of predictions.

    Parameters
    ----------
    uid : str
        Unique identifier for the instance.
    groundtruth : str
        The true label for the instance.
    predictions : list of str
        List of predicted labels.
    scores : list of float
        Confidence scores corresponding to each predicted label.

    Examples
    --------
    >>> classification = Classification(
    ...     uid='123',
    ...     groundtruth='cat',
    ...     predictions=['cat', 'dog', 'bird'],
    ...     scores=[0.9, 0.05, 0.05]
    ... )
    """

    uid: str
    groundtruth: str
    predictions: list[str]
    scores: list[float]

    def __post_init__(self):
        if not isinstance(self.groundtruth, str):
            raise ValueError(
                "A classification must contain a single groundtruth."
            )
        if len(self.predictions) != len(self.scores):
            raise ValueError("There must be a score per prediction label.")



from random import choice, uniform

from valor_lite.classification import Classification, DataLoader


def generate_random_classifications(
    n_classifications: int, n_categories: int, n_labels: int
) -> list[Classification]:

    labels = [str(value) for value in range(n_labels)]

    return [
        Classification(
            uid=f"uid{i}",
            groundtruth=choice(labels),
            predictions=labels,
            scores=[uniform(0, 1) for _ in range(n_labels)],
        )
        for i in range(n_classifications)
    ]


def test_fuzz_classifications():

    quantities = [1, 5, 10]
    for _ in range(100):

        n_classifications = choice(quantities)
        n_categories = choice(quantities)
        n_labels = choice(quantities)

        classifications = generate_random_classifications(
            n_classifications, n_categories=n_categories, n_labels=n_labels
        )

        loader = DataLoader()
        loader.add_data(classifications)
        evaluator = loader.finalize()
        evaluator.evaluate(
            score_thresholds=[0.25, 0.75],
        )


def test_fuzz_classifications_with_filtering():

    quantities = [4, 10]
    for _ in range(100):

        n_classifications = choice(quantities)
        n_categories = choice(quantities)
        n_labels = choice(quantities)

        classifications = generate_random_classifications(
            n_classifications, n_categories=n_categories, n_labels=n_labels
        )

        loader = DataLoader()
        loader.add_data(classifications)
        evaluator = loader.finalize()

        datum_subset = [f"uid{i}" for i in range(len(classifications) // 2)]

        filter_ = evaluator.create_filter(
            datum_uids=datum_subset,
        )

        evaluator.evaluate(
            score_thresholds=[0.25, 0.75],
            filter_=filter_,
        )



import pytest
from valor_lite.text_generation import Metric


def test_metric_type_validation():

    # test type attribute
    with pytest.raises(TypeError):
        Metric(
            type=1234,  # type: ignore - testing
            value=1234,
            parameters={},
        )

    # test value attribute
    with pytest.raises(TypeError):
        Metric(
            type="SomeMetric",
            value=[1, 2, 3],  # type: ignore - testing
            parameters={},
        )

    # test parameters attribute
    with pytest.raises(TypeError):
        Metric(
            type="SomeMetric",
            value=1,
            parameters=123,  # type: ignore - testing
        )

    # test parameter keys
    with pytest.raises(TypeError):
        Metric(
            type="SomeMetric",
            value=1,
            parameters={
                1: "hello",
            },  # type: ignore - testing
        )



from collections import defaultdict
from dataclasses import dataclass

import numpy as np
from numpy.typing import NDArray
from tqdm import tqdm
from valor_lite.semantic_segmentation.annotation import Segmentation
from valor_lite.semantic_segmentation.computation import (
    compute_intermediate_confusion_matrices,
    compute_metrics,
)
from valor_lite.semantic_segmentation.metric import Metric, MetricType
from valor_lite.semantic_segmentation.utilities import (
    unpack_precision_recall_iou_into_metric_lists,
)

"""
Usage
-----

manager = DataLoader()
manager.add_data(
    groundtruths=groundtruths,
    predictions=predictions,
)
evaluator = manager.finalize()

metrics = evaluator.evaluate()

f1_metrics = metrics[MetricType.F1]
accuracy_metrics = metrics[MetricType.Accuracy]

filter_mask = evaluator.create_filter(datum_uids=["uid1", "uid2"])
filtered_metrics = evaluator.evaluate(filter_mask=filter_mask)
"""


@dataclass
class Filter:
    indices: NDArray[np.int32]
    label_metadata: NDArray[np.int32]
    n_pixels: int


class Evaluator:
    """
    Segmentation Evaluator
    """

    def __init__(self):

        # metadata
        self.n_datums = 0
        self.n_groundtruths = 0
        self.n_predictions = 0
        self.n_pixels = 0
        self.n_groundtruth_pixels = 0
        self.n_prediction_pixels = 0
        self.n_labels = 0

        # datum reference
        self.uid_to_index: dict[str, int] = dict()
        self.index_to_uid: dict[int, str] = dict()

        # label reference
        self.label_to_index: dict[str, int] = dict()
        self.index_to_label: dict[int, str] = dict()

        # computation caches
        self._confusion_matrices = np.array([])
        self._label_metadata = np.array([], dtype=np.int32)
        self._label_metadata_per_datum = np.array([], dtype=np.int32)
        self._n_pixels_per_datum = np.array([], dtype=np.int32)

    @property
    def ignored_prediction_labels(self) -> list[str]:
        """
        Prediction labels that are not present in the ground truth set.
        """
        glabels = set(np.where(self._label_metadata[:, 0] > 0)[0])
        plabels = set(np.where(self._label_metadata[:, 1] > 0)[0])
        return [
            self.index_to_label[label_id] for label_id in (plabels - glabels)
        ]

    @property
    def missing_prediction_labels(self) -> list[str]:
        """
        Ground truth labels that are not present in the prediction set.
        """
        glabels = set(np.where(self._label_metadata[:, 0] > 0)[0])
        plabels = set(np.where(self._label_metadata[:, 1] > 0)[0])
        return [
            self.index_to_label[label_id] for label_id in (glabels - plabels)
        ]

    @property
    def metadata(self) -> dict:
        """
        Evaluation metadata.
        """
        return {
            "number_of_datums": self.n_datums,
            "number_of_groundtruths": self.n_groundtruths,
            "number_of_predictions": self.n_predictions,
            "number_of_groundtruth_pixels": self.n_groundtruth_pixels,
            "number_of_prediction_pixels": self.n_prediction_pixels,
            "number_of_labels": self.n_labels,
            "ignored_prediction_labels": self.ignored_prediction_labels,
            "missing_prediction_labels": self.missing_prediction_labels,
        }

    def create_filter(
        self,
        datum_uids: list[str] | NDArray[np.int32] | None = None,
        labels: list[str] | NDArray[np.int32] | None = None,
    ) -> Filter:
        """
        Creates a boolean mask that can be passed to an evaluation.

        Parameters
        ----------
        datum_uids : list[str] | NDArray[np.int32], optional
            An optional list of string uids or a numpy array of uid indices.
        labels : list[tuple[str, str]] | NDArray[np.int32], optional
            An optional list of labels or a numpy array of label indices.

        Returns
        -------
        Filter
            A filter object that can be passed to the `evaluate` method.
        """
        n_datums = self._label_metadata_per_datum.shape[1]
        n_labels = self._label_metadata_per_datum.shape[2]

        mask_datums = np.ones(n_datums, dtype=np.bool_)
        mask_labels = np.ones(n_labels, dtype=np.bool_)

        if datum_uids is not None:
            if isinstance(datum_uids, list):
                datum_uids = np.array(
                    [self.uid_to_index[uid] for uid in datum_uids],
                    dtype=np.int32,
                )
            if datum_uids.size == 0:
                mask_datums[mask_datums] = False
            else:
                mask = (
                    np.arange(n_datums).reshape(-1, 1)
                    == datum_uids.reshape(1, -1)
                ).any(axis=1)
                mask_datums[~mask] = False

        if labels is not None:
            if isinstance(labels, list):
                labels = np.array(
                    [self.label_to_index[label] for label in labels],
                    dtype=np.int32,
                )
            if labels.size == 0:
                mask_labels[mask_labels] = False
            else:
                mask = (
                    np.arange(n_labels).reshape(-1, 1) == labels.reshape(1, -1)
                ).any(axis=1)
                mask_labels[~mask] = False

        mask = mask_datums[:, np.newaxis] & mask_labels[np.newaxis, :]
        label_metadata_per_datum = self._label_metadata_per_datum.copy()
        label_metadata_per_datum[:, ~mask] = 0

        label_metadata = np.zeros_like(self._label_metadata, dtype=np.int32)
        label_metadata = np.transpose(
            np.sum(
                label_metadata_per_datum,
                axis=1,
            )
        )
        n_datums = int(np.sum(label_metadata[:, 0]))

        return Filter(
            indices=np.where(mask_datums)[0],
            label_metadata=label_metadata,
            n_pixels=self._n_pixels_per_datum[mask_datums].sum(),
        )

    def compute_precision_recall_iou(
        self,
        filter_: Filter | None = None,
    ) -> dict[MetricType, list]:
        """
        Performs an evaluation and returns metrics.

        Parameters
        ----------
        filter_ : Filter, optional
            An optional filter object.

        Returns
        -------
        dict[MetricType, list]
            A dictionary mapping MetricType enumerations to lists of computed metrics.
        """

        # apply filters
        data = self._confusion_matrices
        label_metadata = self._label_metadata
        n_pixels = self.n_pixels
        if filter_ is not None:
            data = data[filter_.indices]
            label_metadata = filter_.label_metadata
            n_pixels = filter_.n_pixels

        results = compute_metrics(
            data=data,
            label_metadata=label_metadata,
            n_pixels=n_pixels,
        )

        return unpack_precision_recall_iou_into_metric_lists(
            results=results,
            label_metadata=label_metadata,
            index_to_label=self.index_to_label,
        )

    def evaluate(
        self,
        filter_: Filter | None = None,
    ) -> dict[MetricType, list[Metric]]:
        """
        Computes all available metrics.

        Parameters
        ----------
        filter_ : Filter, optional
            An optional filter object.

        Returns
        -------
        dict[MetricType, list[Metric]]
            Lists of metrics organized by metric type.
        """
        return self.compute_precision_recall_iou(filter_=filter_)


def defaultdict_int():
    return defaultdict(int)


class DataLoader:
    """
    Segmentation DataLoader.
    """

    def __init__(self):
        self._evaluator = Evaluator()
        self.groundtruth_count = defaultdict(defaultdict_int)
        self.prediction_count = defaultdict(defaultdict_int)
        self.matrices = list()
        self.pixel_count = list()

    def _add_datum(self, uid: str) -> int:
        """
        Helper function for adding a datum to the cache.

        Parameters
        ----------
        uid : str
            The datum uid.

        Returns
        -------
        int
            The datum index.
        """
        if uid in self._evaluator.uid_to_index:
            raise ValueError(f"Datum with uid `{uid}` has already been added.")
        index = len(self._evaluator.uid_to_index)
        self._evaluator.uid_to_index[uid] = index
        self._evaluator.index_to_uid[index] = uid
        return index

    def _add_label(self, label: str) -> int:
        """
        Helper function for adding a label to the cache.

        Parameters
        ----------
        label : str
            A string label.

        Returns
        -------
        int
            The label's index.
        """
        if label not in self._evaluator.label_to_index:
            label_id = len(self._evaluator.index_to_label)
            self._evaluator.label_to_index[label] = label_id
            self._evaluator.index_to_label[label_id] = label
        return self._evaluator.label_to_index[label]

    def add_data(
        self,
        segmentations: list[Segmentation],
        show_progress: bool = False,
    ):
        """
        Adds segmentations to the cache.

        Parameters
        ----------
        segmentations : list[Segmentation]
            A list of Segmentation objects.
        show_progress : bool, default=False
            Toggle for tqdm progress bar.
        """

        disable_tqdm = not show_progress
        for segmentation in tqdm(segmentations, disable=disable_tqdm):

            # update metadata
            self._evaluator.n_datums += 1
            self._evaluator.n_groundtruths += len(segmentation.groundtruths)
            self._evaluator.n_predictions += len(segmentation.predictions)
            self._evaluator.n_pixels += segmentation.size
            self._evaluator.n_groundtruth_pixels += segmentation.size * len(
                segmentation.groundtruths
            )
            self._evaluator.n_prediction_pixels += segmentation.size * len(
                segmentation.predictions
            )

            # update datum cache
            uid_index = self._add_datum(segmentation.uid)

            groundtruth_labels = np.full(
                len(segmentation.groundtruths), fill_value=-1
            )
            for idx, groundtruth in enumerate(segmentation.groundtruths):
                label_idx = self._add_label(groundtruth.label)
                groundtruth_labels[idx] = label_idx
                self.groundtruth_count[label_idx][
                    uid_index
                ] += groundtruth.mask.sum()

            prediction_labels = np.full(
                len(segmentation.predictions), fill_value=-1
            )
            for idx, prediction in enumerate(segmentation.predictions):
                label_idx = self._add_label(prediction.label)
                prediction_labels[idx] = label_idx
                self.prediction_count[label_idx][
                    uid_index
                ] += prediction.mask.sum()

            combined_groundtruths = np.stack(
                [
                    groundtruth.mask.flatten()
                    for groundtruth in segmentation.groundtruths
                ],
                axis=0,
            )
            combined_predictions = np.stack(
                [
                    prediction.mask.flatten()
                    for prediction in segmentation.predictions
                ],
                axis=0,
            )

            self.matrices.append(
                compute_intermediate_confusion_matrices(
                    groundtruths=combined_groundtruths,
                    predictions=combined_predictions,
                    groundtruth_labels=groundtruth_labels,
                    prediction_labels=prediction_labels,
                    n_labels=len(self._evaluator.index_to_label),
                )
            )
            self.pixel_count.append(segmentation.size)

    def finalize(self) -> Evaluator:
        """
        Performs data finalization and some preprocessing steps.

        Returns
        -------
        Evaluator
            A ready-to-use evaluator object.
        """

        if len(self.matrices) == 0:
            raise ValueError("No data available to create evaluator.")

        n_datums = self._evaluator.n_datums
        n_labels = len(self._evaluator.index_to_label)

        self._evaluator.n_labels = n_labels

        self._evaluator._label_metadata_per_datum = np.zeros(
            (2, n_datums, n_labels), dtype=np.int32
        )
        for datum_idx in range(n_datums):
            for label_idx in range(n_labels):
                gt_count = (
                    self.groundtruth_count[label_idx].get(datum_idx, 0)
                    if label_idx in self.groundtruth_count
                    else 0
                )
                pd_count = (
                    self.prediction_count[label_idx].get(datum_idx, 0)
                    if label_idx in self.prediction_count
                    else 0
                )
                self._evaluator._label_metadata_per_datum[
                    :, datum_idx, label_idx
                ] = np.array([gt_count, pd_count])

        self._evaluator._label_metadata = np.array(
            [
                [
                    np.sum(
                        self._evaluator._label_metadata_per_datum[
                            0, :, label_idx
                        ]
                    ),
                    np.sum(
                        self._evaluator._label_metadata_per_datum[
                            1, :, label_idx
                        ]
                    ),
                ]
                for label_idx in range(n_labels)
            ],
            dtype=np.int32,
        )

        self._evaluator._n_pixels_per_datum = np.array(
            self.pixel_count, dtype=np.int32
        )

        self._evaluator._confusion_matrices = np.zeros(
            (n_datums, n_labels + 1, n_labels + 1), dtype=np.int32
        )
        for idx, matrix in enumerate(self.matrices):
            h, w = matrix.shape
            self._evaluator._confusion_matrices[idx, :h, :w] = matrix

        return self._evaluator



from collections import defaultdict

import numpy as np
from numpy.typing import NDArray
from valor_lite.object_detection.metric import Metric, MetricType


def unpack_precision_recall_into_metric_lists(
    results: tuple[
        tuple[
            NDArray[np.float64],
            NDArray[np.float64],
            NDArray[np.float64],
            float,
        ],
        tuple[
            NDArray[np.float64],
            NDArray[np.float64],
            NDArray[np.float64],
            float,
        ],
        NDArray[np.float64],
        NDArray[np.float64],
        NDArray[np.float64],
    ],
    iou_thresholds: list[float],
    score_thresholds: list[float],
    index_to_label: dict[int, str],
    label_metadata: NDArray[np.int32],
):
    (
        (
            average_precision,
            mean_average_precision,
            average_precision_average_over_ious,
            mean_average_precision_average_over_ious,
        ),
        (
            average_recall,
            mean_average_recall,
            average_recall_averaged_over_scores,
            mean_average_recall_averaged_over_scores,
        ),
        accuracy,
        precision_recall,
        pr_curves,
    ) = results

    metrics = defaultdict(list)

    metrics[MetricType.AP] = [
        Metric.average_precision(
            value=float(average_precision[iou_idx][label_idx]),
            iou_threshold=iou_threshold,
            label=label,
        )
        for iou_idx, iou_threshold in enumerate(iou_thresholds)
        for label_idx, label in index_to_label.items()
        if int(label_metadata[label_idx, 0]) > 0
    ]

    metrics[MetricType.mAP] = [
        Metric.mean_average_precision(
            value=float(mean_average_precision[iou_idx]),
            iou_threshold=iou_threshold,
        )
        for iou_idx, iou_threshold in enumerate(iou_thresholds)
    ]

    metrics[MetricType.APAveragedOverIOUs] = [
        Metric.average_precision_averaged_over_IOUs(
            value=float(average_precision_average_over_ious[label_idx]),
            iou_thresholds=iou_thresholds,
            label=label,
        )
        for label_idx, label in index_to_label.items()
        if int(label_metadata[label_idx, 0]) > 0
    ]

    metrics[MetricType.mAPAveragedOverIOUs] = [
        Metric.mean_average_precision_averaged_over_IOUs(
            value=float(mean_average_precision_average_over_ious),
            iou_thresholds=iou_thresholds,
        )
    ]

    metrics[MetricType.AR] = [
        Metric.average_recall(
            value=float(average_recall[score_idx, label_idx]),
            iou_thresholds=iou_thresholds,
            score_threshold=score_threshold,
            label=label,
        )
        for score_idx, score_threshold in enumerate(score_thresholds)
        for label_idx, label in index_to_label.items()
        if int(label_metadata[label_idx, 0]) > 0
    ]

    metrics[MetricType.mAR] = [
        Metric.mean_average_recall(
            value=float(mean_average_recall[score_idx]),
            iou_thresholds=iou_thresholds,
            score_threshold=score_threshold,
        )
        for score_idx, score_threshold in enumerate(score_thresholds)
    ]

    metrics[MetricType.ARAveragedOverScores] = [
        Metric.average_recall_averaged_over_scores(
            value=float(average_recall_averaged_over_scores[label_idx]),
            score_thresholds=score_thresholds,
            iou_thresholds=iou_thresholds,
            label=label,
        )
        for label_idx, label in index_to_label.items()
        if int(label_metadata[label_idx, 0]) > 0
    ]

    metrics[MetricType.mARAveragedOverScores] = [
        Metric.mean_average_recall_averaged_over_scores(
            value=float(mean_average_recall_averaged_over_scores),
            score_thresholds=score_thresholds,
            iou_thresholds=iou_thresholds,
        )
    ]

    metrics[MetricType.Accuracy] = [
        Metric.accuracy(
            value=float(accuracy[iou_idx, score_idx]),
            iou_threshold=iou_threshold,
            score_threshold=score_threshold,
        )
        for iou_idx, iou_threshold in enumerate(iou_thresholds)
        for score_idx, score_threshold in enumerate(score_thresholds)
    ]

    metrics[MetricType.PrecisionRecallCurve] = [
        Metric.precision_recall_curve(
            precisions=pr_curves[iou_idx, label_idx, :, 0]
            .astype(float)
            .tolist(),
            scores=pr_curves[iou_idx, label_idx, :, 1].astype(float).tolist(),
            iou_threshold=iou_threshold,
            label=label,
        )
        for iou_idx, iou_threshold in enumerate(iou_thresholds)
        for label_idx, label in index_to_label.items()
        if label_metadata[label_idx, 0] > 0
    ]

    for label_idx, label in index_to_label.items():

        if label_metadata[label_idx, 0] == 0:
            continue

        for score_idx, score_threshold in enumerate(score_thresholds):
            for iou_idx, iou_threshold in enumerate(iou_thresholds):

                row = precision_recall[iou_idx, score_idx, label_idx, :]
                kwargs = {
                    "label": label,
                    "iou_threshold": iou_threshold,
                    "score_threshold": score_threshold,
                }
                metrics[MetricType.Counts].append(
                    Metric.counts(
                        tp=int(row[0]),
                        fp=int(row[1]),
                        fn=int(row[2]),
                        **kwargs,
                    )
                )

                metrics[MetricType.Precision].append(
                    Metric.precision(
                        value=float(row[3]),
                        **kwargs,
                    )
                )
                metrics[MetricType.Recall].append(
                    Metric.recall(
                        value=float(row[4]),
                        **kwargs,
                    )
                )
                metrics[MetricType.F1].append(
                    Metric.f1_score(
                        value=float(row[5]),
                        **kwargs,
                    )
                )

    return metrics


def _convert_example_to_dict(box: NDArray[np.float16]) -> dict[str, float]:
    """
    Converts a cached bounding box example to dictionary format.
    """
    return {
        "xmin": float(box[0]),
        "xmax": float(box[1]),
        "ymin": float(box[2]),
        "ymax": float(box[3]),
    }


def _unpack_confusion_matrix_value(
    confusion_matrix: NDArray[np.float64],
    number_of_labels: int,
    number_of_examples: int,
    index_to_uid: dict[int, str],
    index_to_label: dict[int, str],
    groundtruth_examples: dict[int, NDArray[np.float16]],
    prediction_examples: dict[int, NDArray[np.float16]],
) -> dict[
    str,
    dict[
        str,
        dict[
            str,
            int
            | list[
                dict[
                    str,
                    str | dict[str, float] | float,
                ]
            ],
        ],
    ],
]:
    """
    Unpacks a numpy array of confusion matrix counts and examples.
    """

    datum_idx = lambda gt_label_idx, pd_label_idx, example_idx: int(  # noqa: E731 - lambda fn
        confusion_matrix[
            gt_label_idx,
            pd_label_idx,
            example_idx * 4 + 1,
        ]
    )

    groundtruth_idx = lambda gt_label_idx, pd_label_idx, example_idx: int(  # noqa: E731 - lambda fn
        confusion_matrix[
            gt_label_idx,
            pd_label_idx,
            example_idx * 4 + 2,
        ]
    )

    prediction_idx = lambda gt_label_idx, pd_label_idx, example_idx: int(  # noqa: E731 - lambda fn
        confusion_matrix[
            gt_label_idx,
            pd_label_idx,
            example_idx * 4 + 3,
        ]
    )

    score_idx = lambda gt_label_idx, pd_label_idx, example_idx: float(  # noqa: E731 - lambda fn
        confusion_matrix[
            gt_label_idx,
            pd_label_idx,
            example_idx * 4 + 4,
        ]
    )

    return {
        index_to_label[gt_label_idx]: {
            index_to_label[pd_label_idx]: {
                "count": max(
                    int(confusion_matrix[gt_label_idx, pd_label_idx, 0]),
                    0,
                ),
                "examples": [
                    {
                        "datum": index_to_uid[
                            datum_idx(gt_label_idx, pd_label_idx, example_idx)
                        ],
                        "groundtruth": _convert_example_to_dict(
                            groundtruth_examples[
                                datum_idx(
                                    gt_label_idx,
                                    pd_label_idx,
                                    example_idx,
                                )
                            ][
                                groundtruth_idx(
                                    gt_label_idx,
                                    pd_label_idx,
                                    example_idx,
                                )
                            ]
                        ),
                        "prediction": _convert_example_to_dict(
                            prediction_examples[
                                datum_idx(
                                    gt_label_idx,
                                    pd_label_idx,
                                    example_idx,
                                )
                            ][
                                prediction_idx(
                                    gt_label_idx,
                                    pd_label_idx,
                                    example_idx,
                                )
                            ]
                        ),
                        "score": score_idx(
                            gt_label_idx, pd_label_idx, example_idx
                        ),
                    }
                    for example_idx in range(number_of_examples)
                    if datum_idx(gt_label_idx, pd_label_idx, example_idx) >= 0
                ],
            }
            for pd_label_idx in range(number_of_labels)
        }
        for gt_label_idx in range(number_of_labels)
    }


def _unpack_hallucinations_value(
    hallucinations: NDArray[np.float64],
    number_of_labels: int,
    number_of_examples: int,
    index_to_uid: dict[int, str],
    index_to_label: dict[int, str],
    prediction_examples: dict[int, NDArray[np.float16]],
) -> dict[
    str,
    dict[
        str,
        int | list[dict[str, str | float | dict[str, float]]],
    ],
]:
    """
    Unpacks a numpy array of hallucination counts and examples.
    """

    datum_idx = (
        lambda pd_label_idx, example_idx: int(  # noqa: E731 - lambda fn
            hallucinations[
                pd_label_idx,
                example_idx * 3 + 1,
            ]
        )
    )

    prediction_idx = (
        lambda pd_label_idx, example_idx: int(  # noqa: E731 - lambda fn
            hallucinations[
                pd_label_idx,
                example_idx * 3 + 2,
            ]
        )
    )

    score_idx = (
        lambda pd_label_idx, example_idx: float(  # noqa: E731 - lambda fn
            hallucinations[
                pd_label_idx,
                example_idx * 3 + 3,
            ]
        )
    )

    return {
        index_to_label[pd_label_idx]: {
            "count": max(
                int(hallucinations[pd_label_idx, 0]),
                0,
            ),
            "examples": [
                {
                    "datum": index_to_uid[
                        datum_idx(pd_label_idx, example_idx)
                    ],
                    "prediction": _convert_example_to_dict(
                        prediction_examples[
                            datum_idx(pd_label_idx, example_idx)
                        ][prediction_idx(pd_label_idx, example_idx)]
                    ),
                    "score": score_idx(pd_label_idx, example_idx),
                }
                for example_idx in range(number_of_examples)
                if datum_idx(pd_label_idx, example_idx) >= 0
            ],
        }
        for pd_label_idx in range(number_of_labels)
    }


def _unpack_missing_predictions_value(
    missing_predictions: NDArray[np.int32],
    number_of_labels: int,
    number_of_examples: int,
    index_to_uid: dict[int, str],
    index_to_label: dict[int, str],
    groundtruth_examples: dict[int, NDArray[np.float16]],
) -> dict[str, dict[str, int | list[dict[str, str | dict[str, float]]]]]:
    """
    Unpacks a numpy array of missing prediction counts and examples.
    """

    datum_idx = (
        lambda gt_label_idx, example_idx: int(  # noqa: E731 - lambda fn
            missing_predictions[
                gt_label_idx,
                example_idx * 2 + 1,
            ]
        )
    )

    groundtruth_idx = (
        lambda gt_label_idx, example_idx: int(  # noqa: E731 - lambda fn
            missing_predictions[
                gt_label_idx,
                example_idx * 2 + 2,
            ]
        )
    )

    return {
        index_to_label[gt_label_idx]: {
            "count": max(
                int(missing_predictions[gt_label_idx, 0]),
                0,
            ),
            "examples": [
                {
                    "datum": index_to_uid[
                        datum_idx(gt_label_idx, example_idx)
                    ],
                    "groundtruth": _convert_example_to_dict(
                        groundtruth_examples[
                            datum_idx(gt_label_idx, example_idx)
                        ][groundtruth_idx(gt_label_idx, example_idx)]
                    ),
                }
                for example_idx in range(number_of_examples)
                if datum_idx(gt_label_idx, example_idx) >= 0
            ],
        }
        for gt_label_idx in range(number_of_labels)
    }


def unpack_confusion_matrix_into_metric_list(
    results: tuple[
        NDArray[np.float64],
        NDArray[np.float64],
        NDArray[np.int32],
    ],
    iou_thresholds: list[float],
    score_thresholds: list[float],
    number_of_examples: int,
    index_to_label: dict[int, str],
    index_to_uid: dict[int, str],
    groundtruth_examples: dict[int, NDArray[np.float16]],
    prediction_examples: dict[int, NDArray[np.float16]],
) -> list[Metric]:
    (
        confusion_matrix,
        hallucinations,
        missing_predictions,
    ) = results
    n_labels = len(index_to_label)
    return [
        Metric.confusion_matrix(
            iou_threshold=iou_threshold,
            score_threshold=score_threshold,
            maximum_number_of_examples=number_of_examples,
            confusion_matrix=_unpack_confusion_matrix_value(
                confusion_matrix=confusion_matrix[iou_idx, score_idx, :, :, :],
                number_of_labels=n_labels,
                number_of_examples=number_of_examples,
                index_to_label=index_to_label,
                index_to_uid=index_to_uid,
                groundtruth_examples=groundtruth_examples,
                prediction_examples=prediction_examples,
            ),
            hallucinations=_unpack_hallucinations_value(
                hallucinations=hallucinations[iou_idx, score_idx, :, :],
                number_of_labels=n_labels,
                number_of_examples=number_of_examples,
                index_to_label=index_to_label,
                index_to_uid=index_to_uid,
                prediction_examples=prediction_examples,
            ),
            missing_predictions=_unpack_missing_predictions_value(
                missing_predictions=missing_predictions[
                    iou_idx, score_idx, :, :
                ],
                number_of_labels=n_labels,
                number_of_examples=number_of_examples,
                index_to_label=index_to_label,
                index_to_uid=index_to_uid,
                groundtruth_examples=groundtruth_examples,
            ),
        )
        for iou_idx, iou_threshold in enumerate(iou_thresholds)
        for score_idx, score_threshold in enumerate(score_thresholds)
    ]



import numpy as np
from valor_lite.classification import Classification, DataLoader, Metric


def test_metadata_using_classification_example(
    classifications_animal_example: list[Classification],
):
    manager = DataLoader()
    manager.add_data(classifications_animal_example)
    evaluator = manager.finalize()

    assert evaluator.ignored_prediction_labels == []
    assert evaluator.missing_prediction_labels == []
    assert evaluator.n_datums == 6
    assert evaluator.n_labels == 3
    assert evaluator.n_groundtruths == 6
    assert evaluator.n_predictions == 3 * 6

    assert evaluator.metadata == {
        "n_datums": 6,
        "n_groundtruths": 6,
        "n_predictions": 3 * 6,
        "n_labels": 3,
        "ignored_prediction_labels": [],
        "missing_prediction_labels": [],
    }


def _flatten_metrics(m) -> list:
    if isinstance(m, dict):
        keys = list(m.keys())
        values = [
            inner_value
            for value in m.values()
            for inner_value in _flatten_metrics(value)
        ]
        return keys + values
    elif isinstance(m, list):
        return [
            inner_value
            for value in m
            for inner_value in _flatten_metrics(value)
        ]
    elif isinstance(m, Metric):
        return _flatten_metrics(m.to_dict())
    else:
        return [m]


def test_output_types_dont_contain_numpy(
    basic_classifications: list[Classification],
):
    manager = DataLoader()
    manager.add_data(basic_classifications)
    evaluator = manager.finalize()

    metrics = evaluator.evaluate(
        score_thresholds=[0.25, 0.75],
    )

    values = _flatten_metrics(metrics)
    for value in values:
        if isinstance(value, (np.generic, np.ndarray, Metric)):
            raise TypeError(value)



import numpy as np
from valor_lite.classification import (
    Classification,
    DataLoader,
    MetricType,
    compute_precision_recall_rocauc,
)


def test_accuracy_computation():

    # groundtruth, prediction, score
    data = np.array(
        [
            # datum 0
            [0, 0, 0, 1.0, 1],  # tp
            [0, 0, 1, 0.0, 0],  # tn
            [0, 0, 2, 0.0, 0],  # tn
            [0, 0, 3, 0.0, 0],  # tn
            # datum 1
            [1, 0, 0, 0.0, 0],  # fn
            [1, 0, 1, 0.0, 0],  # tn
            [1, 0, 2, 1.0, 1],  # fp
            [1, 0, 3, 0.0, 0],  # tn
            # datum 2
            [2, 3, 0, 0.0, 0],  # tn
            [2, 3, 1, 0.0, 0],  # tn
            [2, 3, 2, 0.0, 0],  # tn
            [2, 3, 3, 0.3, 1],  # fn for score threshold > 0.3
        ],
        dtype=np.float64,
    )

    # groundtruth count, prediction count, label key
    label_metadata = np.array(
        [
            [2, 1, 0],
            [0, 2, 0],
            [1, 0, 0],
            [1, 1, 0],
        ],
        dtype=np.int32,
    )

    score_thresholds = np.array([0.25, 0.75], dtype=np.float64)

    (_, _, _, accuracy, _, _, _) = compute_precision_recall_rocauc(
        data=data,
        label_metadata=label_metadata,
        score_thresholds=score_thresholds,
        n_datums=3,
        hardmax=False,
    )

    # score threshold, label, count metric
    assert accuracy.shape == (2,)

    # score >= 0.25
    assert accuracy[0] == 2 / 3
    # score >= 0.75
    assert accuracy[1] == 1 / 3


def test_accuracy_basic(basic_classifications: list[Classification]):
    loader = DataLoader()
    loader.add_data(basic_classifications)
    evaluator = loader.finalize()

    assert evaluator.metadata == {
        "n_datums": 3,
        "n_groundtruths": 3,
        "n_predictions": 12,
        "n_labels": 4,
        "ignored_prediction_labels": ["1", "2"],
        "missing_prediction_labels": [],
    }

    metrics = evaluator.evaluate(score_thresholds=[0.25, 0.75])

    actual_metrics = [m.to_dict() for m in metrics[MetricType.Accuracy]]
    expected_metrics = [
        {
            "type": "Accuracy",
            "value": 2 / 3,
            "parameters": {
                "score_threshold": 0.25,
                "hardmax": True,
            },
        },
        {
            "type": "Accuracy",
            "value": 1 / 3,
            "parameters": {
                "score_threshold": 0.75,
                "hardmax": True,
            },
        },
    ]
    for m in actual_metrics:
        assert m in expected_metrics
    for m in expected_metrics:
        assert m in actual_metrics


def test_accuracy_with_animal_example(
    classifications_animal_example: list[Classification],
):

    loader = DataLoader()
    loader.add_data(classifications_animal_example)
    evaluator = loader.finalize()

    metrics = evaluator.evaluate(score_thresholds=[0.5])

    actual_metrics = [m.to_dict() for m in metrics[MetricType.Accuracy]]
    expected_metrics = [
        {
            "type": "Accuracy",
            "value": 2 / 6,
            "parameters": {
                "score_threshold": 0.5,
                "hardmax": True,
            },
        },
    ]
    for m in actual_metrics:
        assert m in expected_metrics
    for m in expected_metrics:
        assert m in actual_metrics


def test_accuracy_color_example(
    classifications_color_example: list[Classification],
):

    loader = DataLoader()
    loader.add_data(classifications_color_example)
    evaluator = loader.finalize()

    metrics = evaluator.evaluate(score_thresholds=[0.5])

    actual_metrics = [m.to_dict() for m in metrics[MetricType.Accuracy]]
    expected_metrics = [
        {
            "type": "Accuracy",
            "value": 2 / 6,
            "parameters": {
                "score_threshold": 0.5,
                "hardmax": True,
            },
        },
    ]
    for m in actual_metrics:
        assert m in expected_metrics
    for m in expected_metrics:
        assert m in actual_metrics


def test_accuracy_with_image_example(
    classifications_image_example: list[Classification],
):
    loader = DataLoader()
    loader.add_data(classifications_image_example)
    evaluator = loader.finalize()

    assert evaluator.metadata == {
        "n_datums": 2,
        "n_groundtruths": 2,
        "n_predictions": 4,
        "n_labels": 4,
        "ignored_prediction_labels": ["v1", "v8", "v5"],
        "missing_prediction_labels": [],
    }

    metrics = evaluator.evaluate()

    actual_metrics = [m.to_dict() for m in metrics[MetricType.Accuracy]]
    expected_metrics = [
        {
            "type": "Accuracy",
            "value": 0.5,
            "parameters": {
                "score_threshold": 0.0,
                "hardmax": True,
            },
        },
    ]
    for m in actual_metrics:
        assert m in expected_metrics
    for m in expected_metrics:
        assert m in actual_metrics


def test_accuracy_with_tabular_example(
    classifications_tabular_example: list[Classification],
):
    loader = DataLoader()
    loader.add_data(classifications_tabular_example)
    evaluator = loader.finalize()

    assert evaluator.metadata == {
        "n_datums": 10,
        "n_groundtruths": 10,
        "n_predictions": 30,
        "n_labels": 3,
        "ignored_prediction_labels": [],
        "missing_prediction_labels": [],
    }

    metrics = evaluator.evaluate()

    actual_metrics = [m.to_dict() for m in metrics[MetricType.Accuracy]]
    expected_metrics = [
        {
            "type": "Accuracy",
            "value": 0.5,
            "parameters": {
                "score_threshold": 0.0,
                "hardmax": True,
            },
        },
    ]
    for m in actual_metrics:
        assert m in expected_metrics
    for m in expected_metrics:
        assert m in actual_metrics



import json
import re
from typing import Any

from valor_lite.text_generation.llm.exceptions import InvalidLLMResponseError


def trim_and_load_json(text: str) -> dict[str, Any]:
    """
    Trims and loads input_string as a dictionary. Adapted from DeepEval https://github.com/confident-ai/deepeval/blob/dc117a5ea2160dbb61909c537908a41f7da4dfe7/deepeval/metrics/utils.py#L50

    Parameters
    ----------
    input_string : str
        The input string to trim and load as a json.

    Returns
    -------
    dict
        A dictionary.
    """

    pattern = r"\{[\s\S]*\}"
    match = re.search(pattern, text)
    if not match:
        raise InvalidLLMResponseError(
            f"LLM did not include valid brackets in its response: {text}"
        )
    extracted_text = match.group()

    try:
        return json.loads(extracted_text)
    except json.JSONDecodeError as e:
        raise InvalidLLMResponseError(
            f"Evaluation LLM responded with invalid JSON. JSONDecodeError: {str(e)}"
        )


def find_first_signed_integer(text: str) -> int | None:
    match = re.search(r"-?\d+", text)
    if not match:
        return None
    return int(match.group())



import json

import pytest
from valor_lite.text_generation.llm.exceptions import InvalidLLMResponseError
from valor_lite.text_generation.llm.utilities import (
    find_first_signed_integer,
    trim_and_load_json,
)


def test_trim_and_load_json():

    # test removal of excess text
    expected = {
        "verdicts": [
            {"verdict": "yes"},
            {
                "verdict": "no",
                "reason": "The statement 'I also think puppies are cute.' is irrelevant to the question about who the cutest cat ever is.",
            },
        ]
    }
    input_string = (
        "this text should be trimmed"
        + json.dumps(expected, indent=4)
        + "also should be trimmed"
    )
    assert trim_and_load_json(input_string) == expected

    # test missing starting bracket
    input = """
        "sentence": "Hello, world!",
        "value": 3
    }"""
    with pytest.raises(InvalidLLMResponseError) as e:
        trim_and_load_json(input)
    assert "LLM did not include valid brackets in its response" in str(e)

    # test missing closing bracket a '}'
    input = '{"field": "value" '
    with pytest.raises(InvalidLLMResponseError) as e:
        trim_and_load_json(input)
    assert "LLM did not include valid brackets in its response" in str(e)

    # test missing comma edge case
    input = """{
        "verdicts": [
            {
                "verdict": "yes"
            }
            {
                "verdict": "no",
                "reason": "The statement 'I also think puppies are cute.' is irrelevant to the question about who the cutest cat ever is."
            }
        ]
    }"""
    with pytest.raises(InvalidLLMResponseError) as e:
        trim_and_load_json(input)
    assert "Evaluation LLM responded with invalid JSON." in str(e)

    # test dictionary with non-string keys
    input = "{1: 'a'}"
    with pytest.raises(InvalidLLMResponseError) as e:
        trim_and_load_json(input)
    assert "Evaluation LLM responded with invalid JSON." in str(e)


def test_find_first_signed_integer():

    text = "hello world the number is -101."
    assert find_first_signed_integer(text) == -101

    text = "hidden number 314 in text."
    assert find_first_signed_integer(text) == 314

    text = "only return first number 1, 2, 3"
    assert find_first_signed_integer(text) == 1

    text = "return nothing if no integers"
    assert find_first_signed_integer(text) is None

    text = "floating values are not registered 0.1"
    assert find_first_signed_integer(text) == 0



import datetime
import os
from unittest.mock import MagicMock

import pytest
from valor_lite.text_generation.llm.integrations import (
    MistralWrapper,
    OpenAIWrapper,
    _validate_messages,
)

try:
    import mistralai
    from mistralai.models import (
        AssistantMessage,
        ChatCompletionChoice,
        ChatCompletionResponse,
        UsageInfo,
    )
    from mistralai.models.sdkerror import SDKError as MistralSDKError
except ImportError:
    mistralai = None

try:
    import openai
    from openai import OpenAIError
    from openai.types.chat import ChatCompletionMessage
    from openai.types.chat.chat_completion import ChatCompletion, Choice
    from openai.types.completion_usage import CompletionUsage
except ImportError:
    openai = None


def test__validate_messages():
    # Valid messages.
    _validate_messages(
        messages=[
            {"role": "system", "content": "You are a helpful assistant."},
            {"role": "user", "content": "Hello!"},
        ],
    )

    # messages must be a list
    with pytest.raises(TypeError):
        _validate_messages(
            messages={"role": "system", "content": "You are a helpful assistant."}  # type: ignore - testing
        )

    # messages must be a list of dictionaries
    with pytest.raises(TypeError):
        _validate_messages(
            messages=["You are a helpful assistant."]  # type: ignore - testing
        )

    # Each message must have 'role' and 'content' keys
    with pytest.raises(ValueError):
        _validate_messages(
            messages=[
                {"role": "system", "content": "You are a helpful assistant."},
                {"role": "user"},
            ]
        )

    # Role values must be strings
    with pytest.raises(TypeError):
        _validate_messages(
            messages=[
                {"role": "system", "content": "You are a helpful assistant."},
                {"role": 1, "content": "Hello!"},  # type: ignore - testing
            ]
        )

    # Content values must be a string
    with pytest.raises(TypeError):
        _validate_messages(
            messages=[
                {"role": "system", "content": "You are a helpful assistant."},
                {"role": "user", "content": 1},  # type: ignore - testing
            ]
        )


@pytest.mark.skipif(
    openai is None,
    reason="Openai is not installed.",
)
def test_openai_client():
    def _create_bad_request(model, messages, seed):
        raise ValueError

    def _create_mock_chat_completion_with_bad_length(
        model, messages, seed
    ) -> ChatCompletion:  # type: ignore - test is not run if openai is not installed
        return ChatCompletion(  # type: ignore - test is not run if openai is not installed
            id="foo",
            model="gpt-3.5-turbo",
            object="chat.completion",
            choices=[
                Choice(  # type: ignore - test is not run if openai is not installed
                    finish_reason="length",
                    index=0,
                    message=ChatCompletionMessage(  # type: ignore - test is not run if openai is not installed
                        content="some response",
                        role="assistant",
                    ),
                )
            ],
            usage=CompletionUsage(  # type: ignore - test is not run if openai is not installed
                completion_tokens=1, prompt_tokens=2, total_tokens=3
            ),
            created=int(datetime.datetime.now().timestamp()),
        )

    def _create_mock_chat_completion_with_content_filter(
        model, messages, seed
    ) -> ChatCompletion:  # type: ignore - test is not run if openai is not installed
        return ChatCompletion(  # type: ignore - test is not run if openai is not installed
            id="foo",
            model="gpt-3.5-turbo",
            object="chat.completion",
            choices=[
                Choice(  # type: ignore - test is not run if openai is not installed
                    finish_reason="content_filter",
                    index=0,
                    message=ChatCompletionMessage(  # type: ignore - test is not run if openai is not installed
                        content="some response",
                        role="assistant",
                    ),
                )
            ],
            usage=CompletionUsage(  # type: ignore - test is not run if openai is not installed
                completion_tokens=1, prompt_tokens=2, total_tokens=3
            ),
            created=int(datetime.datetime.now().timestamp()),
        )

    def _create_mock_chat_completion(model, messages, seed) -> ChatCompletion:  # type: ignore - test is not run if openai is not installed
        return ChatCompletion(  # type: ignore - test is not run if openai is not installed
            id="foo",
            model="gpt-3.5-turbo",
            object="chat.completion",
            choices=[
                Choice(  # type: ignore - test is not run if openai is not installed
                    finish_reason="stop",
                    index=0,
                    message=ChatCompletionMessage(  # type: ignore - test is not run if openai is not installed
                        content="some response",
                        role="assistant",
                    ),
                )
            ],
            usage=CompletionUsage(  # type: ignore - test is not run if openai is not installed
                completion_tokens=1, prompt_tokens=2, total_tokens=3
            ),
            created=int(datetime.datetime.now().timestamp()),
        )

    def _create_mock_chat_completion_none_content(
        model, messages, seed
    ) -> ChatCompletion:  # type: ignore - test is not run if openai is not installed
        return ChatCompletion(  # type: ignore - test is not run if openai is not installed
            id="foo",
            model="gpt-3.5-turbo",
            object="chat.completion",
            choices=[
                Choice(  # type: ignore - test is not run if openai is not installed
                    finish_reason="stop",
                    index=0,
                    message=ChatCompletionMessage(  # type: ignore - test is not run if openai is not installed
                        content=None,
                        role="assistant",
                    ),
                )
            ],
            usage=CompletionUsage(  # type: ignore - test is not run if openai is not installed
                completion_tokens=1, prompt_tokens=2, total_tokens=3
            ),
            created=int(datetime.datetime.now().timestamp()),
        )

    # OpenAI client call should fail as the API key is invalid.
    client = OpenAIWrapper(api_key="invalid_key", model_name="model_name")
    fake_message = [
        {"role": "system", "content": "You are a helpful assistant."}
    ]
    with pytest.raises(OpenAIError):  # type: ignore - test is not run if openai is not installed
        client(fake_message)

    # The OpenAI Client should be able to connect if the API key is set as the environment variable.
    os.environ["OPENAI_API_KEY"] = "dummy_key"
    client = OpenAIWrapper(model_name="model_name")

    client.client = MagicMock()

    # A bad request should raise a ValueError.
    client.client.chat.completions.create = _create_bad_request
    with pytest.raises(ValueError) as e:
        client(fake_message)

    # The metric computation should fail when the finish reason is bad length.
    client.client.chat.completions.create = (
        _create_mock_chat_completion_with_bad_length
    )
    with pytest.raises(ValueError) as e:
        client(fake_message)
    assert "reached max token limit" in str(e)

    # The metric computation should fail when the finish reason is content filter.
    client.client.chat.completions.create = (
        _create_mock_chat_completion_with_content_filter
    )
    with pytest.raises(ValueError) as e:
        client(fake_message)
    assert "flagged by content filter" in str(e)

    # Should run successfully when the finish reason is stop.
    client.client.chat.completions.create = _create_mock_chat_completion
    assert client(fake_message) == "some response"

    # Should run successfully even when the response content is None.
    client.client.chat.completions.create = (
        _create_mock_chat_completion_none_content
    )
    assert client(fake_message) == ""


@pytest.mark.skipif(
    mistralai is None,
    reason="MistralAI is not installed.",
)
def test_mistral_client():
    def _create_bad_request(model, messages):
        raise ValueError

    def _create_mock_chat_completion_with_bad_length(
        model,
        messages,
    ) -> ChatCompletionResponse:  # type: ignore - test is not run if mistralai is not installed
        return ChatCompletionResponse(  # type: ignore - test is not run if mistralai is not installed
            id="foo",
            model="gpt-3.5-turbo",
            object="chat.completion",
            choices=[
                ChatCompletionChoice(  # type: ignore - test is not run if mistralai is not installed
                    finish_reason="length",
                    index=0,
                    message=AssistantMessage(  # type: ignore - test is not run if mistralai is not installed
                        role="assistant",
                        content="some response",
                        name=None,  # type: ignore - mistralai issue
                        tool_calls=None,
                        tool_call_id=None,  # type: ignore - mistralai issue
                    ),
                )
            ],
            created=int(datetime.datetime.now().timestamp()),
            usage=UsageInfo(  # type: ignore - test is not run if mistralai is not installed
                prompt_tokens=2, total_tokens=4, completion_tokens=199
            ),
        )

    def _create_mock_chat_completion(
        model, messages
    ) -> ChatCompletionResponse:  # type: ignore - test is not run if mistralai is not installed
        return ChatCompletionResponse(  # type: ignore - test is not run if mistralai is not installed
            id="foo",
            model="gpt-3.5-turbo",
            object="chat.completion",
            choices=[
                ChatCompletionChoice(  # type: ignore - test is not run if mistralai is not installed
                    finish_reason="stop",
                    index=0,
                    message=AssistantMessage(  # type: ignore - test is not run if mistralai is not installed
                        role="assistant",
                        content="some response",
                        name=None,  # type: ignore - mistralai issue
                        tool_calls=None,
                        tool_call_id=None,  # type: ignore - mistralai issue
                    ),
                )
            ],
            created=int(datetime.datetime.now().timestamp()),
            usage=UsageInfo(  # type: ignore - test is not run if mistralai is not installed
                prompt_tokens=2, total_tokens=4, completion_tokens=199
            ),
        )

    def _create_mock_chat_completion_choices_is_None(
        model, messages
    ) -> ChatCompletionResponse:  # type: ignore - test is not run if mistralai is not installed
        return ChatCompletionResponse(  # type: ignore - test is not run if mistralai is not installed
            id="foo",
            model="gpt-3.5-turbo",
            object="chat.completion",
            choices=None,
            created=int(datetime.datetime.now().timestamp()),
            usage=UsageInfo(  # type: ignore - test is not run if mistralai is not installed
                prompt_tokens=2, total_tokens=4, completion_tokens=199
            ),
        )

    def _create_mock_chat_completion_bad_message_content(
        model, messages
    ) -> ChatCompletionResponse:  # type: ignore - test is not run if mistralai is not installed
        return ChatCompletionResponse(  # type: ignore - test is not run if mistralai is not installed
            id="foo",
            model="gpt-3.5-turbo",
            object="chat.completion",
            choices=[
                ChatCompletionChoice(  # type: ignore - test is not run if mistralai is not installed
                    finish_reason="stop",
                    index=0,
                    message=AssistantMessage(  # type: ignore - test is not run if mistralai is not installed
                        role="assistant",
                        name=None,  # type: ignore - mistralai issue
                        tool_calls=None,
                        tool_call_id=None,  # type: ignore - mistralai issue
                    ),
                )
            ],
            created=int(datetime.datetime.now().timestamp()),
            usage=UsageInfo(  # type: ignore - test is not run if mistralai is not installed
                prompt_tokens=2, total_tokens=4, completion_tokens=199
            ),
        )

    # Mistral client call should fail as the API key is invalid.
    client = MistralWrapper(api_key="invalid_key", model_name="model_name")
    fake_message = [{"role": "assistant", "content": "content"}]
    with pytest.raises(MistralSDKError):  # type: ignore - test is not run if mistralai is not installed
        client(fake_message)

    # The Mistral Client should be able to connect if the API key is set as the environment variable.
    os.environ["MISTRAL_API_KEY"] = "dummy_key"
    client = MistralWrapper(model_name="model_name")

    client.client = MagicMock()

    # The metric computation should fail if the request fails.
    client.client.chat.complete = _create_bad_request
    with pytest.raises(ValueError) as e:
        client(fake_message)

    # The metric computation should fail when the finish reason is bad length.
    client.client.chat.complete = _create_mock_chat_completion_with_bad_length
    with pytest.raises(ValueError) as e:
        client(fake_message)
    assert "reached max token limit" in str(e)

    # The metric computation should run successfully when the finish reason is stop.
    client.client.chat.complete = _create_mock_chat_completion
    assert client(fake_message) == "some response"

    # The metric computation should run successfully even when choices is None.
    client.client.chat.complete = _create_mock_chat_completion_choices_is_None
    assert client(fake_message) == ""

    # The metric computation should fail when the message doesn't contain any content.
    client.client.chat.complete = (
        _create_mock_chat_completion_bad_message_content
    )
    with pytest.raises(TypeError) as e:
        client(fake_message)
        assert "Mistral AI response was not a string." in str(e)



import numpy as np
from valor_lite.semantic_segmentation import (
    Bitmask,
    DataLoader,
    MetricType,
    Segmentation,
)


def test_confusion_matrix_basic_segmentations(
    basic_segmentations: list[Segmentation],
):
    loader = DataLoader()
    loader.add_data(basic_segmentations)
    evaluator = loader.finalize()

    metrics = evaluator.evaluate()

    actual_metrics = [m.to_dict() for m in metrics[MetricType.ConfusionMatrix]]
    expected_metrics = [
        {
            "type": "ConfusionMatrix",
            "value": {
                "confusion_matrix": {
                    "v1": {"v1": {"iou": 0.5}, "v2": {"iou": 0.0}},
                    "v2": {"v1": {"iou": 0.0}, "v2": {"iou": 0.5}},
                },
                "hallucinations": {
                    "v1": {"ratio": 0.0},
                    "v2": {"ratio": 0.5},
                },
                "missing_predictions": {
                    "v1": {"ratio": 0.5},
                    "v2": {"ratio": 0.0},
                },
            },
            "parameters": {},
        },
    ]
    for m in actual_metrics:
        assert m in expected_metrics
    for m in expected_metrics:
        assert m in actual_metrics


def test_confusion_matrix_segmentations_from_boxes(
    segmentations_from_boxes: list[Segmentation],
):
    loader = DataLoader()
    loader.add_data(segmentations_from_boxes)
    evaluator = loader.finalize()

    metrics = evaluator.evaluate()

    actual_metrics = [m.to_dict() for m in metrics[MetricType.ConfusionMatrix]]
    expected_metrics = [
        {
            "type": "ConfusionMatrix",
            "value": {
                "confusion_matrix": {
                    "v1": {
                        "v1": {
                            "iou": 5000 / (10000 + 10000 - 5000)
                        },  # 50% overlap
                        "v2": {"iou": 0.0},
                    },
                    "v2": {
                        "v1": {"iou": 0.0},
                        "v2": {
                            "iou": 1 / (14999 + 4999 + 1)  # overlaps 1 pixel
                        },
                    },
                },
                "hallucinations": {
                    "v1": {"ratio": 5000 / 10000},  # 50% overlap
                    "v2": {
                        "ratio": 4999 / 5000
                    },  # overlaps 1 pixel out of 5000 predictions
                },
                "missing_predictions": {
                    "v1": {"ratio": 5000 / 10000},
                    "v2": {
                        "ratio": 14999 / 15000
                    },  # overlaps 1 pixel out of 15,000 groundtruths
                },
            },
            "parameters": {},
        },
    ]
    for m in actual_metrics:
        assert m in expected_metrics
    for m in expected_metrics:
        assert m in actual_metrics


def test_confusion_matrix_intermediate_counting():

    segmentation = Segmentation(
        uid="uid1",
        groundtruths=[
            Bitmask(
                mask=np.array([[False, False], [True, False]]),
                label="a",
            ),
            Bitmask(
                mask=np.array([[False, False], [False, True]]),
                label="b",
            ),
            Bitmask(
                mask=np.array([[True, False], [False, False]]),
                label="c",
            ),
            Bitmask(
                mask=np.array([[False, True], [False, False]]),
                label="d",
            ),
        ],
        predictions=[
            Bitmask(
                mask=np.array([[False, False], [False, False]]),
                label="a",
            ),
            Bitmask(
                mask=np.array([[False, False], [False, False]]),
                label="b",
            ),
            Bitmask(
                mask=np.array([[True, True], [True, True]]),
                label="c",
            ),
            Bitmask(
                mask=np.array([[False, False], [False, False]]),
                label="d",
            ),
        ],
    )

    loader = DataLoader()
    loader.add_data([segmentation])

    assert len(loader.matrices) == 1
    assert (
        loader.matrices[0]
        == np.array(
            [
                [0, 0, 0, 0, 0],
                [0, 0, 0, 1, 0],
                [0, 0, 0, 1, 0],
                [0, 0, 0, 1, 0],
                [0, 0, 0, 1, 0],
            ]
        )
    ).all()



from random import choice, uniform

from valor_lite.object_detection import BoundingBox, DataLoader, Detection


def _generate_random_detections(
    n_detections: int, n_boxes: int, labels: str
) -> list[Detection]:
    def bbox(is_prediction):
        width, height = 50, 50
        xmin, ymin = uniform(0, 1000), uniform(0, 1000)
        xmax, ymax = uniform(xmin, xmin + width), uniform(ymin, ymin + height)
        kw = {"scores": [uniform(0, 1)]} if is_prediction else {}
        return BoundingBox(
            xmin,
            xmax,
            ymin,
            ymax,
            [choice(labels)],
            **kw,
        )

    return [
        Detection(
            uid=f"uid{i}",
            groundtruths=[bbox(is_prediction=False) for _ in range(n_boxes)],
            predictions=[bbox(is_prediction=True) for _ in range(n_boxes)],
        )
        for i in range(n_detections)
    ]


def test_fuzz_detections():

    few_labels = "abc"
    many_labels = "abcdefghijklmnopqrstuvwxyz123456789"
    quantities = [1, 5, 10]

    for _ in range(100):

        labels = choice([few_labels, many_labels])
        n_detections = choice(quantities)
        n_boxes = choice(quantities)

        detections = _generate_random_detections(n_detections, n_boxes, labels)

        loader = DataLoader()
        loader.add_bounding_boxes(detections)
        evaluator = loader.finalize()
        evaluator.evaluate(
            iou_thresholds=[0.25, 0.75],
            score_thresholds=[0.25, 0.75],
        )


def test_fuzz_detections_with_filtering():

    few_labels = "abcd"
    many_labels = "abcdefghijklmnopqrstuvwxyz123456789"
    quantities = [4, 10]

    for _ in range(100):

        labels = choice([few_labels, many_labels])
        n_detections = choice(quantities)
        n_boxes = choice(quantities)

        detections = _generate_random_detections(n_detections, n_boxes, labels)

        loader = DataLoader()
        loader.add_bounding_boxes(detections)
        evaluator = loader.finalize()

        datum_subset = [f"uid{i}" for i in range(len(detections) // 2)]

        filter_ = evaluator.create_filter(
            datum_uids=datum_subset,
        )

        evaluator.evaluate(
            iou_thresholds=[0.25, 0.75],
            score_thresholds=[0.25, 0.75],
            filter_=filter_,
        )


def test_fuzz_confusion_matrix():
    dets = _generate_random_detections(1000, 30, "abcde")
    loader = DataLoader()
    loader.add_bounding_boxes(dets)
    evaluator = loader.finalize()
    assert evaluator.metadata == {
        "ignored_prediction_labels": [],
        "missing_prediction_labels": [],
        "n_datums": 1000,
        "n_groundtruths": 30000,
        "n_predictions": 30000,
        "n_labels": 5,
    }
    evaluator.evaluate(
        iou_thresholds=[0.25, 0.75],
        score_thresholds=[0.5],
    )



from valor_lite.semantic_segmentation import (
    DataLoader,
    Metric,
    MetricType,
    Segmentation,
)


def test_recall_basic_segmentations(basic_segmentations: list[Segmentation]):
    loader = DataLoader()
    loader.add_data(basic_segmentations)
    evaluator = loader.finalize()

    metrics = evaluator.evaluate()

    actual_metrics = [m.to_dict() for m in metrics[MetricType.Recall]]
    expected_metrics = [
        {
            "type": "Recall",
            "value": 0.5,
            "parameters": {"label": "v1"},
        },
        {
            "type": "Recall",
            "value": 1.0,
            "parameters": {"label": "v2"},
        },
    ]
    for m in actual_metrics:
        assert m in expected_metrics
    for m in expected_metrics:
        assert m in actual_metrics


def test_recall_segmentations_from_boxes(
    segmentations_from_boxes: list[Segmentation],
):
    loader = DataLoader()
    loader.add_data(segmentations_from_boxes)
    evaluator = loader.finalize()

    metrics = evaluator.evaluate()

    actual_metrics = [m.to_dict() for m in metrics[MetricType.Recall]]
    expected_metrics = [
        {
            "type": "Recall",
            "value": 0.5,  # 50% overlap
            "parameters": {"label": "v1"},
        },
        {
            "type": "Recall",
            "value": 1 / 15000,  # overlaps 1 pixel out of 15,000 groundtruths
            "parameters": {"label": "v2"},
        },
    ]
    for m in actual_metrics:
        assert m in expected_metrics
    for m in expected_metrics:
        assert m in actual_metrics


def test_recall_large_random_segmentations(
    large_random_segmentations: list[Segmentation],
):
    loader = DataLoader()
    loader.add_data(large_random_segmentations)
    evaluator = loader.finalize()

    metrics = evaluator.evaluate()

    for m in metrics[MetricType.Recall]:
        assert isinstance(m, Metric)
        match m.parameters["label"]:
            case "v1":
                assert round(m.value, 1) == 0.9  # type: ignore - testing
            case "v2":
                assert round(m.value, 2) == 0.09  # type: ignore - testing
            case "v3":
                assert round(m.value, 2) == 0.01  # type: ignore - testing
            case "v4":
                assert round(m.value, 1) == 0.4  # type: ignore - testing
            case "v5":
                assert round(m.value, 1) == 0.4  # type: ignore - testing
            case "v6":
                assert round(m.value, 1) == 0.1  # type: ignore - testing
            case "v7":
                assert round(m.value, 1) == 0.3  # type: ignore - testing
            case "v8":
                assert round(m.value, 1) == 0.3  # type: ignore - testing
            case "v9":
                assert round(m.value, 1) == 0.3  # type: ignore - testing
            case _:
                assert False



import json
import os
from dataclasses import dataclass
from datetime import datetime
from enum import Enum
from pathlib import Path
from time import time

import requests
from tqdm import tqdm
from valor_lite.object_detection import BoundingBox, DataLoader, Detection


def _get_bbox_extrema(
    data: list,
) -> tuple[float, float, float, float]:
    """Get the bounding box coordinates from a valor Annotation object."""
    x = [point[0] for shape in data for point in shape]
    y = [point[1] for shape in data for point in shape]
    return (min(x), max(x), min(y), max(y))


def _convert_valor_dicts_into_Detection(gt_dict: dict, pd_dict: dict):
    """Convert a groundtruth dictionary and prediction dictionary into a valor_lite Detection object."""
    gts = []
    pds = []

    for gann in gt_dict["annotations"]:
        labels = []
        for valor_label in gann["labels"]:
            # NOTE: we only include labels where the key is "name"
            if valor_label["key"] != "name":
                continue

            labels.append(valor_label["value"])

        # if the annotation doesn't contain any labels that aren't key == 'name', then we skip that annotation
        if not labels:
            continue

        x_min, x_max, y_min, y_max = _get_bbox_extrema(gann["bounding_box"])

        gts.append(
            BoundingBox(
                xmin=x_min,
                xmax=x_max,
                ymin=y_min,
                ymax=y_max,
                labels=labels,
            )
        )

    for pann in pd_dict["annotations"]:
        labels, scores = [], []
        for valor_label in pann["labels"]:
            if valor_label["key"] != "name":
                continue
            labels.append(valor_label["value"])
            scores.append(valor_label["score"])
        if not labels:
            continue

        x_min, x_max, y_min, y_max = _get_bbox_extrema(pann["bounding_box"])

        pds.append(
            BoundingBox(
                xmin=x_min,
                xmax=x_max,
                ymin=y_min,
                ymax=y_max,
                labels=labels,
                scores=scores,
            )
        )

    return Detection(
        uid=gt_dict["datum"]["uid"],
        groundtruths=gts,
        predictions=pds,
    )


class AnnotationType(str, Enum):
    NONE = "none"
    BOX = "box"
    POLYGON = "polygon"
    MULTIPOLYGON = "multipolygon"
    RASTER = "raster"


def time_it(fn):
    def wrapper(*args, **kwargs):
        start = time()
        results = fn(*args, **kwargs)
        return (time() - start, results)

    return wrapper


def download_data_if_not_exists(
    file_name: str,
    file_path: Path,
    url: str,
):
    """Download the data from a public bucket if it doesn't exist locally."""

    if not os.path.exists(file_path):
        response = requests.get(url, stream=True)
        if response.status_code == 200:
            total_size = int(response.headers.get("content-length", 0))
            with open(file_path, "wb") as f:
                with tqdm(
                    total=total_size,
                    unit="B",
                    unit_scale=True,
                    unit_divisor=1024,
                    desc=file_name,
                ) as pbar:
                    for chunk in response.iter_content(chunk_size=1024):
                        if chunk:
                            f.write(chunk)
                            pbar.update(1024)
        else:
            raise RuntimeError(response)
    else:
        print(f"{file_name} already exists locally.")

    # sort file by datum uid
    with open(file_path, "r") as f:
        lines = [x for x in f]
    with open(file_path, "w") as f:
        for line in sorted(
            lines, key=lambda x: int(json.loads(x)["datum"]["uid"])
        ):
            f.write(line)


def write_results_to_file(write_path: Path, results: list[dict]):
    """Write results to json"""
    current_datetime = datetime.now().strftime("%d/%m/%Y %H:%M:%S")
    if os.path.isfile(write_path):
        with open(write_path, "r") as file:
            file.seek(0)
            data = json.load(file)
    else:
        data = {}

    data[current_datetime] = results

    with open(write_path, "w+") as file:
        json.dump(data, file, indent=4)


@time_it
def ingest(
    manager: DataLoader,
    gt_path: Path,
    pd_path: Path,
    limit: int,
    chunk_size: int,
):
    accumulated_time = 0.0
    with open(gt_path, "r") as gf:
        with open(pd_path, "r") as pf:

            count = 0
            detections = []
            for gline, pline in zip(gf, pf):

                gt_dict = json.loads(gline)
                pd_dict = json.loads(pline)
                detections.append(
                    _convert_valor_dicts_into_Detection(
                        gt_dict=gt_dict, pd_dict=pd_dict
                    )
                )

                count += 1
                if count >= limit and limit > 0:
                    break
                elif len(detections) < chunk_size or chunk_size == -1:
                    continue

                timer, _ = time_it(manager.add_bounding_boxes)(detections)
                accumulated_time += timer
                detections = []

            if detections:
                timer, _ = time_it(manager.add_bounding_boxes)(detections)
                accumulated_time += timer

    return accumulated_time


@dataclass
class Benchmark:
    limit: int
    n_datums: int
    n_groundtruths: int
    n_predictions: int
    n_labels: int
    gt_type: AnnotationType
    pd_type: AnnotationType
    chunk_size: int
    ingestion: float
    preprocessing: float
    precomputation: float
    evaluation: float
    detailed_curves: list[tuple[int, float]]

    def result(self) -> dict:
        return {
            "limit": self.limit,
            "n_datums": self.n_datums,
            "n_groundtruths": self.n_groundtruths,
            "n_predictions": self.n_predictions,
            "n_labels": self.n_labels,
            "dtype": {
                "groundtruth": self.gt_type.value,
                "prediction": self.pd_type.value,
            },
            "chunk_size": self.chunk_size,
            "ingestion": {
                "loading_from_file": f"{round(self.ingestion - self.preprocessing, 2)} seconds",
                "numpy_conversion + IoU": f"{round(self.preprocessing, 2)} seconds",
                "ranking_pairs": f"{round(self.precomputation, 2)} seconds",
                "total": f"{round(self.ingestion + self.precomputation, 2)} seconds",
            },
            "base_evaluation": f"{round(self.evaluation, 2)} seconds",
            "detailed_evaluation": [
                {
                    "n_points": 10,
                    "n_examples": curve[0],
                    "computation": f"{round(curve[1], 2)} seconds",
                }
                for curve in self.detailed_curves
            ],
        }


def run_benchmarking_analysis(
    limits_to_test: list[int],
    combinations: list[tuple[AnnotationType, AnnotationType]] | None = None,
    results_file: str = "objdet_results.json",
    chunk_size: int = -1,
    compute_pr: bool = True,
    compute_detailed: bool = True,
    ingestion_timeout=30,
    evaluation_timeout=30,
):
    """Time various function calls and export the results."""
    current_directory = Path(__file__).parent
    write_path = current_directory / Path(results_file)

    gt_box_filename = "gt_objdet_coco_bbox.jsonl"
    gt_polygon_filename = "gt_objdet_coco_polygon.jsonl"
    gt_multipolygon_filename = "gt_objdet_coco_raster_multipolygon.jsonl"
    gt_raster_filename = "gt_objdet_coco_raster_bitmask.jsonl"
    pd_box_filename = "pd_objdet_yolo_bbox.jsonl"
    pd_polygon_filename = "pd_objdet_yolo_polygon.jsonl"
    pd_multipolygon_filename = "pd_objdet_yolo_multipolygon.jsonl"
    pd_raster_filename = "pd_objdet_yolo_raster.jsonl"

    groundtruth_caches = {
        AnnotationType.BOX: gt_box_filename,
        AnnotationType.POLYGON: gt_polygon_filename,
        AnnotationType.MULTIPOLYGON: gt_multipolygon_filename,
        AnnotationType.RASTER: gt_raster_filename,
    }
    prediction_caches = {
        AnnotationType.BOX: pd_box_filename,
        AnnotationType.POLYGON: pd_polygon_filename,
        AnnotationType.MULTIPOLYGON: pd_multipolygon_filename,
        AnnotationType.RASTER: pd_raster_filename,
    }

    # default is to perform all combinations
    if combinations is None:
        combinations = [
            (gt_type, pd_type)
            for gt_type in groundtruth_caches
            for pd_type in prediction_caches
        ]

    # cache data locally
    filenames = [
        *list(groundtruth_caches.values()),
        *list(prediction_caches.values()),
    ]
    for filename in filenames:
        file_path = current_directory / Path(filename)
        url = f"https://pub-fae71003f78140bdaedf32a7c8d331d2.r2.dev/{filename}"
        download_data_if_not_exists(
            file_name=filename, file_path=file_path, url=url
        )

    # iterate through datum limits
    results = list()
    for limit in limits_to_test:
        for gt_type, pd_type in combinations:

            gt_filename = groundtruth_caches[gt_type]
            pd_filename = prediction_caches[pd_type]

            # === Base Evaluation ===
            manager = DataLoader()

            # ingest + preprocess
            (ingest_time, preprocessing_time,) = ingest(
                manager=manager,
                gt_path=current_directory / Path(gt_filename),
                pd_path=current_directory / Path(pd_filename),
                limit=limit,
                chunk_size=chunk_size,
            )  # type: ignore - time_it wrapper

            finalization_time, evaluator = time_it(manager.finalize)()

            if ingest_time > ingestion_timeout and ingestion_timeout != -1:
                raise TimeoutError(
                    f"Base precomputation timed out with limit of {limit}."
                )

            # evaluate - base metrics only
            eval_time, metrics = time_it(evaluator.compute_precision_recall)()
            if eval_time > evaluation_timeout and evaluation_timeout != -1:
                raise TimeoutError(
                    f"Base evaluation timed out with {evaluator.n_datums} datums."
                )

            # evaluate - base metrics + detailed counts with no samples
            detailed_counts_time_no_samples, metrics = time_it(
                evaluator.compute_confusion_matrix
            )(
                number_of_examples=0,
            )
            if (
                detailed_counts_time_no_samples > evaluation_timeout
                and evaluation_timeout != -1
            ):
                raise TimeoutError(
                    f"Detailed evaluation w/ no samples timed out with {evaluator.n_datums} datums."
                )

            # evaluate - base metrics + detailed counts with 3 samples
            detailed_counts_time_three_samples, metrics = time_it(
                evaluator.compute_confusion_matrix
            )(
                number_of_examples=3,
            )
            if (
                detailed_counts_time_three_samples > evaluation_timeout
                and evaluation_timeout != -1
            ):
                raise TimeoutError(
                    f"Detailed w/ 3 samples evaluation timed out with {evaluator.n_datums} datums."
                )

            results.append(
                Benchmark(
                    limit=limit,
                    n_datums=evaluator.n_datums,
                    n_groundtruths=evaluator.n_groundtruths,
                    n_predictions=evaluator.n_predictions,
                    n_labels=evaluator.n_labels,
                    gt_type=gt_type,
                    pd_type=pd_type,
                    chunk_size=chunk_size,
                    ingestion=ingest_time,
                    preprocessing=preprocessing_time,
                    precomputation=finalization_time,
                    evaluation=eval_time,
                    detailed_curves=[
                        (0, detailed_counts_time_no_samples),
                        (3, detailed_counts_time_three_samples),
                    ],
                ).result()
            )

    write_results_to_file(write_path=write_path, results=results)


if __name__ == "__main__":

    # run bounding box benchmark
    run_benchmarking_analysis(
        combinations=[
            (AnnotationType.BOX, AnnotationType.BOX),
        ],
        limits_to_test=[5000, 5000],
        compute_detailed=False,
    )

    # # run polygon benchmark
    # run_benchmarking_analysis(
    #     combinations=[
    #         (AnnotationType.POLYGON, AnnotationType.POLYGON),
    #     ],
    #     limits_to_test=[5000, 5000],
    #     compute_detailed=False,
    # )

    # # run raster benchmark
    # run_benchmarking_analysis(
    #     combinations=[
    #         (AnnotationType.RASTER, AnnotationType.RASTER),
    #     ],
    #     limits_to_test=[500, 500],
    #     compute_detailed=False,
    # )



from valor_lite.profiling import create_runtime_profiler
from valor_lite.semantic_segmentation import DataLoader, generate_segmentation


def benchmark_add_data(
    n_labels: int,
    shape: tuple[int, int],
    time_limit: float | None,
    repeat: int = 1,
) -> float:
    """
    Benchmarks 'Dataloader.add_data' for semantic segmentation.

    Parameters
    ----------
    n_labels : int
        The number of unique labels to generate.
    shape : tuple[int, int]
        The size (h,w) of the mask to generate.
    time_limit : float, optional
        An optional time limit to constrain the benchmark.
    repeat : int
        The number of times to run the benchmark to produce a runtime average.

    Returns
    -------
    float
        The average runtime.
    """

    profile = create_runtime_profiler(
        time_limit=time_limit,
        repeat=repeat,
    )

    elapsed = 0
    for _ in range(repeat):
        data = generate_segmentation(
            datum_uid="uid",
            number_of_unique_labels=n_labels,
            mask_height=shape[0],
            mask_width=shape[1],
        )
        loader = DataLoader()
        elapsed += profile(loader.add_data)([data])
    return elapsed / repeat


def benchmark_finalize(
    n_datums: int,
    n_labels: int,
    time_limit: float | None,
    repeat: int = 1,
):
    """
    Benchmarks 'Dataloader.finalize' for semantic segmentation.

    Parameters
    ----------
    n_datums : int
        The number of datums to generate.
    n_labels : int
        The number of unique labels to generate.
    time_limit : float, optional
        An optional time limit to constrain the benchmark.
    repeat : int
        The number of times to run the benchmark to produce a runtime average.

    Returns
    -------
    float
        The average runtime.
    """

    profile = create_runtime_profiler(
        time_limit=time_limit,
        repeat=repeat,
    )

    elapsed = 0
    for _ in range(repeat):

        data = [
            generate_segmentation(
                datum_uid=str(i),
                number_of_unique_labels=n_labels,
                mask_height=5,
                mask_width=5,
            )
            for i in range(10)
        ]
        loader = DataLoader()
        for datum_idx in range(n_datums):
            segmentation = data[datum_idx % 10]
            segmentation.uid = str(datum_idx)
            loader.add_data([segmentation])
        elapsed += profile(loader.finalize)()
    return elapsed / repeat


def benchmark_evaluate(
    n_datums: int,
    n_labels: int,
    time_limit: float | None,
    repeat: int = 1,
):
    """
    Benchmarks 'Evaluator.evaluate' for semantic segmentation.

    Parameters
    ----------
    n_datums : int
        The number of datums to generate.
    n_labels : int
        The number of unique labels to generate.
    time_limit : float, optional
        An optional time limit to constrain the benchmark.
    repeat : int
        The number of times to run the benchmark to produce a runtime average.

    Returns
    -------
    float
        The average runtime.
    """

    profile = create_runtime_profiler(
        time_limit=time_limit,
        repeat=repeat,
    )

    elapsed = 0
    for _ in range(repeat):

        data = [
            generate_segmentation(
                datum_uid=str(i),
                number_of_unique_labels=n_labels,
                mask_height=5,
                mask_width=5,
            )
            for i in range(10)
        ]
        loader = DataLoader()
        for datum_idx in range(n_datums):
            segmentation = data[datum_idx % 10]
            segmentation.uid = str(datum_idx)
            loader.add_data([segmentation])
        evaluator = loader.finalize()
        elapsed += profile(evaluator.evaluate)()
    return elapsed / repeat



import pytest
from valor_lite.classification import Classification


@pytest.fixture
def basic_classifications() -> list[Classification]:
    return [
        Classification(
            uid="uid0",
            groundtruth="0",
            predictions=[
                "0",
                "1",
                "2",
                "3",
            ],
            scores=[1.0, 0.0, 0.0, 0.0],
        ),
        Classification(
            uid="uid1",
            groundtruth="0",
            predictions=[
                "0",
                "1",
                "2",
                "3",
            ],
            scores=[0.0, 0.0, 1.0, 0.0],
        ),
        Classification(
            uid="uid2",
            groundtruth="3",
            predictions=[
                "0",
                "1",
                "2",
                "3",
            ],
            scores=[0.0, 0.0, 0.0, 0.3],
        ),
    ]


@pytest.fixture
def classifications_from_api_unit_tests() -> list[Classification]:
    """c.f. with

    ```
    from sklearn.metrics import classification_report

    y_true = [0, 0, 0, 1, 2, 2]
    y_pred = [0, 1, 2, 1, 1, 1]

    print(classification_report(y_true, y_pred))
    ```
    """
    return [
        Classification(
            uid="uid0",
            groundtruth="0",
            predictions=[
                "0",
                "1",
                "2",
            ],
            scores=[1.0, 0.0, 0.0],
        ),
        Classification(
            uid="uid1",
            groundtruth="0",
            predictions=[
                "0",
                "1",
                "2",
            ],
            scores=[0.0, 1.0, 0.0],
        ),
        Classification(
            uid="uid2",
            groundtruth="0",
            predictions=[
                "0",
                "1",
                "2",
            ],
            scores=[0.0, 0.0, 1.0],
        ),
        Classification(
            uid="uid3",
            groundtruth="1",
            predictions=[
                "0",
                "1",
                "2",
            ],
            scores=[0.0, 1.0, 0.0],
        ),
        Classification(
            uid="uid4",
            groundtruth="2",
            predictions=[
                "0",
                "1",
                "2",
            ],
            scores=[0.0, 1.0, 0.0],
        ),
        Classification(
            uid="uid5",
            groundtruth="2",
            predictions=[
                "0",
                "1",
                "2",
            ],
            scores=[0.0, 1.0, 0.0],
        ),
    ]


@pytest.fixture
def classifications_animal_example() -> list[Classification]:
    animal_gts = ["bird", "dog", "bird", "bird", "cat", "dog"]
    animal_pds = [
        {"bird": 0.6, "dog": 0.2, "cat": 0.2},
        {"cat": 0.9, "dog": 0.1, "bird": 0.0},
        {"cat": 0.8, "dog": 0.05, "bird": 0.15},
        {"dog": 0.75, "cat": 0.1, "bird": 0.15},
        {"cat": 1.0, "dog": 0.0, "bird": 0.0},
        {"cat": 0.4, "dog": 0.4, "bird": 0.2},
        # Note: In the case of a tied score, the ordering of predictions is used.
    ]

    return [
        Classification(
            uid=f"uid{idx}",
            groundtruth=gt,
            predictions=list(pd.keys()),
            scores=list(pd.values()),
        )
        for idx, (gt, pd) in enumerate(zip(animal_gts, animal_pds))
    ]


@pytest.fixture
def classifications_color_example() -> list[Classification]:
    color_gts = ["white", "white", "red", "blue", "black", "red"]
    color_pds = [
        {"white": 0.65, "red": 0.1, "blue": 0.2, "black": 0.05},
        {"blue": 0.5, "white": 0.3, "red": 0.0, "black": 0.2},
        {"red": 0.4, "white": 0.2, "blue": 0.1, "black": 0.3},
        {"white": 1.0, "red": 0.0, "blue": 0.0, "black": 0.0},
        {"red": 0.8, "white": 0.0, "blue": 0.2, "black": 0.0},
        {"red": 0.9, "white": 0.06, "blue": 0.01, "black": 0.03},
    ]

    return [
        Classification(
            uid=f"uid{idx}",
            groundtruth=gt,
            predictions=list(pd.keys()),
            scores=list(pd.values()),
        )
        for idx, (gt, pd) in enumerate(zip(color_gts, color_pds))
    ]


@pytest.fixture
def classifications_image_example() -> list[Classification]:
    return [
        Classification(
            uid="uid5",
            groundtruth="v4",
            predictions=[
                "v1",
                "v8",
            ],
            scores=[0.47, 0.53],
        ),
        Classification(
            uid="uid6",
            groundtruth="v4",
            predictions=[
                "v4",
                "v5",
            ],
            scores=[0.71, 0.29],
        ),
    ]


@pytest.fixture
def classifications_tabular_example() -> list[Classification]:
    gt_clfs_tabular = [1, 1, 2, 0, 0, 0, 1, 1, 1, 1]
    pred_clfs_tabular = [
        [0.37, 0.35, 0.28],
        [0.24, 0.61, 0.15],
        [0.03, 0.88, 0.09],
        [0.97, 0.03, 0.0],
        [1.0, 0.0, 0.0],
        [1.0, 0.0, 0.0],
        [0.01, 0.96, 0.03],
        [0.28, 0.02, 0.7],
        [0.78, 0.21, 0.01],
        [0.45, 0.11, 0.44],
    ]
    return [
        Classification(
            uid=f"uid{i}",
            groundtruth=str(gt_label),
            predictions=[str(pd_label) for pd_label, _ in enumerate(pds)],
            scores=pds,
        )
        for i, (gt_label, pds) in enumerate(
            zip(gt_clfs_tabular, pred_clfs_tabular)
        )
    ]


@pytest.fixture
def classifications_no_predictions() -> list[Classification]:
    return [
        Classification(
            uid="uid1",
            groundtruth="v1",
            predictions=[],
            scores=[],
        )
    ]


@pytest.fixture
def classifications_multiclass() -> list[Classification]:
    return [
        Classification(
            uid="uid0",
            groundtruth="cat",
            predictions=[
                "cat",
                "dog",
                "bee",
            ],
            scores=[
                0.44598543489942505,
                0.3255517969601126,
                0.22846276814046224,
            ],
        ),
        Classification(
            uid="uid1",
            groundtruth="bee",
            predictions=[
                "cat",
                "dog",
                "bee",
            ],
            scores=[
                0.4076893257212283,
                0.14780458563955237,
                0.4445060886392194,
            ],
        ),
        Classification(
            uid="uid2",
            groundtruth="cat",
            predictions=[
                "cat",
                "dog",
                "bee",
            ],
            scores=[
                0.25060075263871917,
                0.3467428086425673,
                0.4026564387187136,
            ],
        ),
        Classification(
            uid="uid3",
            groundtruth="bee",
            predictions=[
                "cat",
                "dog",
                "bee",
            ],
            scores=[
                0.2003514145616792,
                0.2485912151889644,
                0.5510573702493565,
            ],
        ),
        Classification(
            uid="uid4",
            groundtruth="dog",
            predictions=[
                "cat",
                "dog",
                "bee",
            ],
            scores=[
                0.33443897813714385,
                0.5890646197236098,
                0.07649640213924616,
            ],
        ),
    ]


@pytest.fixture
def classifications_multiclass_true_negatives_check() -> (
    list[Classification]
):
    return [
        Classification(
            uid="uid1",
            groundtruth="ant",
            predictions=["ant", "bee", "cat"],
            scores=[0.15, 0.48, 0.37],
        ),
    ]


@pytest.fixture
def classifications_multiclass_zero_count() -> list[Classification]:
    return [
        Classification(
            uid="uid1",
            groundtruth="ant",
            predictions=["ant", "bee", "cat"],
            scores=[0.15, 0.48, 0.37],
        )
    ]



from dataclasses import dataclass, field

import numpy as np
from numpy.typing import NDArray


@dataclass
class Bitmask:
    """
    Represents a binary mask with an associated semantic label.

    Parameters
    ----------
    mask : NDArray[np.bool_]
        A NumPy array of boolean values representing the mask.
    label : str
        The semantic label associated with the mask.

    Examples
    --------
    >>> import numpy as np
    >>> mask = np.array([[True, False], [False, True]], dtype=np.bool_)
    >>> bitmask = Bitmask(mask=mask, label='ocean')
    """

    mask: NDArray[np.bool_]
    label: str

    def __post_init__(self):
        if self.mask.dtype != np.bool_:
            raise ValueError(
                f"Bitmask recieved mask with dtype '{self.mask.dtype}'."
            )


@dataclass
class Segmentation:
    """
    Segmentation data structure holding ground truth and prediction bitmasks for semantic segmentation tasks.

    Parameters
    ----------
    uid : str
        Unique identifier for the image or sample.
    groundtruths : List[Bitmask]
        List of ground truth bitmasks.
    predictions : List[Bitmask]
        List of predicted bitmasks.
    shape : tuple of int, optional
        The shape of the segmentation masks. This is set automatically after initialization.
    size : int, optional
        The total number of pixels in the masks. This is set automatically after initialization.

    Examples
    --------
    >>> import numpy as np
    >>> mask1 = np.array([[True, False], [False, True]], dtype=np.bool_)
    >>> groundtruth = Bitmask(mask=mask1, label='object')
    >>> mask2 = np.array([[False, True], [True, False]], dtype=np.bool_)
    >>> prediction = Bitmask(mask=mask2, label='object')
    >>> segmentation = Segmentation(
    ...     uid='123',
    ...     groundtruths=[groundtruth],
    ...     predictions=[prediction]
    ... )
    """

    uid: str
    groundtruths: list[Bitmask]
    predictions: list[Bitmask]
    shape: tuple[int, ...] = field(default_factory=lambda: (0, 0))
    size: int = field(default=0)

    def __post_init__(self):

        groundtruth_shape = {
            groundtruth.mask.shape for groundtruth in self.groundtruths
        }
        prediction_shape = {
            prediction.mask.shape for prediction in self.predictions
        }
        if len(groundtruth_shape) == 0:
            raise ValueError("The segmenation is missing ground truths.")
        elif len(prediction_shape) == 0:
            raise ValueError("The segmenation is missing predictions.")
        elif (
            len(groundtruth_shape) != 1
            or len(prediction_shape) != 1
            or groundtruth_shape != prediction_shape
        ):
            raise ValueError(
                "A shape mismatch exists within the segmentation."
            )

        self.shape = groundtruth_shape.pop()
        self.size = int(np.prod(np.array(self.shape)))


def generate_segmentation(
    datum_uid: str,
    number_of_unique_labels: int,
    mask_height: int,
    mask_width: int,
) -> Segmentation:
    """
    Generates a semantic segmentation annotation.

    Parameters
    ----------
    datum_uid : str
        The datum UID for the generated segmentation.
    number_of_unique_labels : int
        The number of unique labels.
    mask_height : int
        The height of the mask in pixels.
    mask_width : int
        The width of the mask in pixels.

    Returns
    -------
    Segmentation
        A generated semantic segmenatation annotation.
    """

    if number_of_unique_labels > 1:
        common_proba = 0.4 / (number_of_unique_labels - 1)
        min_proba = min(common_proba, 0.1)
        labels = [str(i) for i in range(number_of_unique_labels)] + [None]
        proba = (
            [0.5]
            + [common_proba for _ in range(number_of_unique_labels - 1)]
            + [0.1]
        )
    elif number_of_unique_labels == 1:
        labels = ["0", None]
        proba = [0.9, 0.1]
        min_proba = 0.1
    else:
        raise ValueError(
            "The number of unique labels should be greater than zero."
        )

    probabilities = np.array(proba, dtype=np.float64)
    weights = (probabilities / min_proba).astype(np.int32)

    indices = np.random.choice(
        np.arange(len(weights)),
        size=(mask_height * 2, mask_width),
        p=probabilities,
    )

    N = len(labels)

    masks = np.arange(N)[:, None, None] == indices

    gts = []
    pds = []
    for lidx in range(N):
        label = labels[lidx]
        if label is None:
            continue
        gts.append(
            Bitmask(
                mask=masks[lidx, :mask_height, :],
                label=label,
            )
        )
        pds.append(
            Bitmask(
                mask=masks[lidx, mask_height:, :],
                label=label,
            )
        )

    return Segmentation(
        uid=datum_uid,
        groundtruths=gts,
        predictions=pds,
    )



import pytest
from valor_lite.text_generation import Evaluator, QueryResponse
from valor_lite.text_generation.computation import calculate_summary_coherence
from valor_lite.text_generation.llm.exceptions import InvalidLLMResponseError


def test_calculate_summary_coherence(mock_client):

    evaluator = Evaluator(client=mock_client)

    for i in [1, 2, 3, 4, 5]:
        mock_client.returning = str(i)
        assert (
            calculate_summary_coherence(
                client=mock_client,
                system_prompt="",
                text="a",
                summary="b",
            )
            == i
        )
        assert evaluator.compute_summary_coherence(
            response=QueryResponse(
                query="a",
                response="b",
            )
        ).to_dict() == {
            "type": "SummaryCoherence",
            "value": i,
            "parameters": {
                "evaluator": "mock",
                "retries": 0,
            },
        }

    for i in [-1, 0, 6]:
        mock_client.returning = str(i)
        with pytest.raises(InvalidLLMResponseError):
            calculate_summary_coherence(
                client=mock_client,
                system_prompt="",
                text="a",
                summary="b",
            )

        assert evaluator.compute_summary_coherence(
            response=QueryResponse(
                query="a",
                response="b",
            )
        ).to_dict() == {
            "type": "Error",
            "value": {
                "type": "InvalidLLMResponseError",
                "message": f"Summary coherence score was not an integer between 1 and 5: {i}",
            },
            "parameters": {
                "evaluator": "mock",
                "retries": 0,
            },
        }

    for i in ["one", "three"]:
        mock_client.returning = str(i)
        with pytest.raises(InvalidLLMResponseError):
            calculate_summary_coherence(
                client=mock_client,
                system_prompt="",
                text="a",
                summary="b",
            )

        assert evaluator.compute_summary_coherence(
            response=QueryResponse(
                query="a",
                response="b",
            )
        ).to_dict() == {
            "type": "Error",
            "value": {
                "type": "InvalidLLMResponseError",
                "message": f"LLM response was not a valid summary coherence score: {i}",
            },
            "parameters": {
                "evaluator": "mock",
                "retries": 0,
            },
        }



import numpy as np
from valor_lite.classification import Classification, DataLoader, MetricType
from valor_lite.classification.computation import (
    compute_precision_recall_rocauc,
)


def test_compute_rocauc_animals():
    """
    Test ROC auc computation. This agrees with scikit-learn: the code (whose data
    comes from classification_test_data)

    animal_gts = ["bird", "dog", "bird", "bird", "cat", "dog"]
    animal_preds = [
        {"bird": 0.6, "dog": 0.2, "cat": 0.2},
        {"cat": 0.9, "dog": 0.1, "bird": 0.0},
        {"cat": 0.8, "dog": 0.05, "bird": 0.15},
        {"dog": 0.75, "cat": 0.1, "bird": 0.15},
        {"cat": 1.0, "dog": 0.0, "bird": 0.0},
        {"cat": 0.4, "dog": 0.4, "bird": 0.2},
    ]

    ```
    from sklearn.metrics import roc_auc_score

    # for the "animal" label key
    y_true = [0, 2, 0, 0, 1, 2]
    y_score = [
        [0.6, 0.2, 0.2],
        [0.0, 0.9, 0.1],
        [0.15, 0.8, 0.05],
        [0.15, 0.1, 0.75],
        [0.0, 1.0, 0.0],
        [0.2, 0.4, 0.4],
    ]

    print(roc_auc_score(y_true, y_score, multi_class="ovr"))

    ```

    outputs:

    ```
    0.8009259259259259
    ```
    """

    # groundtruth, prediction, score
    animals = np.array(
        [
            # (animal, bird)
            [0, 0, 0, 0.6, 0],
            [0, 1, 0, 0.2, 0],
            [0, 0, 0, 0.15, 0],
            [0, 0, 0, 0.15, 0],
            [0, 1, 0, 0.0, 0],
            [0, 2, 0, 0.0, 0],
            # (animal, dog)
            [0, 0, 1, 0.75, 0],
            [0, 1, 1, 0.4, 0],
            [0, 0, 1, 0.2, 0],
            [0, 1, 1, 0.1, 0],
            [0, 0, 1, 0.05, 0],
            [0, 2, 1, 0.0, 0],
            # (animal, cat)
            [0, 2, 2, 1.0, 0],
            [0, 1, 2, 0.9, 0],
            [0, 0, 2, 0.8, 0],
            [0, 1, 2, 0.4, 0],
            [0, 0, 2, 0.2, 0],
            [0, 0, 2, 0.1, 0],
        ],
        dtype=np.float64,
    )
    indices = np.argsort(-animals[:, 3], axis=0)
    animals = animals[indices]

    # create args
    label_metadata = np.array(
        [
            [3, 6],
            [2, 6],
            [1, 6],
        ],
        dtype=np.int32,
    )

    # compute ROCAUC and mROCAUC
    (_, _, _, _, _, rocauc, mean_rocauc) = compute_precision_recall_rocauc(
        data=animals,
        label_metadata=label_metadata,
        n_datums=6,
        score_thresholds=np.array([]),
        hardmax=False,
    )

    # test ROCAUC
    assert rocauc.shape == (label_metadata.shape[0],)
    assert rocauc[0] == 0.7777777777777778  # (animal, bird)
    assert rocauc[1] == 0.625  # (animal, dog)
    assert rocauc[2] == 1.0  # (animal, cat)

    # test mROCAUC
    assert mean_rocauc == 0.8009259259259259


def test_compute_rocauc_colors():
    """
    Test ROC auc computation. This agrees with scikit-learn: the code (whose data
    comes from classification_test_data)

    color_gts = ["white", "white", "red", "blue", "black", "red"]
    color_preds = [
        {"white": 0.65, "red": 0.1, "blue": 0.2, "black": 0.05},
        {"blue": 0.5, "white": 0.3, "red": 0.0, "black": 0.2},
        {"red": 0.4, "white": 0.2, "blue": 0.1, "black": 0.3},
        {"white": 1.0, "red": 0.0, "blue": 0.0, "black": 0.0},
        {"red": 0.8, "white": 0.0, "blue": 0.2, "black": 0.0},
        {"red": 0.9, "white": 0.06, "blue": 0.01, "black": 0.03},
    ]

    ```
    from sklearn.metrics import roc_auc_score

    # for the "color" label key
    y_true = [3, 0, 2, 1, 0, 2]
    y_score = [
        [0.05, 0.2, 0.1, 0.65],
        [0.2, 0.5, 0.0, 0.3],
        [0.3, 0.1, 0.4, 0.2],
        [0.0, 0.0, 0.0, 1.0],
        [0.0, 0.2, 0.8, 0.0],
        [0.03, 0.01, 0.9, 0.06],
    ]
    ```

    outputs:

    ```
    0.43125
    ```
    """

    # groundtruth, prediction, score
    colors = np.array(
        [
            [3, 2, 0, 1, 1],
            [5, 1, 1, 0.9, 1],
            [4, 3, 1, 0.8, 1],
            [0, 0, 0, 0.65, 1],
            [1, 0, 2, 0.5, 1],
            [2, 1, 1, 0.4, 1],
            [1, 0, 0, 0.3, 0],
            [2, 1, 3, 0.3, 0],
            [2, 1, 0, 0.2, 0],
            [0, 0, 2, 0.2, 0],
            [4, 3, 2, 0.2, 0],
            [1, 0, 3, 0.2, 0],
            [0, 0, 1, 0.1, 0],
            [2, 1, 2, 0.1, 0],
            [5, 1, 0, 0.06, 0],
            [0, 0, 3, 0.05, 0],
            [5, 1, 3, 0.03, 0],
            [5, 1, 2, 0.01, 0],
            [4, 3, 0, 0, 0],
            [1, 0, 1, 0, 0],
            [3, 2, 1, 0, 0],
            [3, 2, 2, 0, 0],
            [3, 2, 3, 0, 0],
            [4, 3, 3, 0, 0],
        ],
        dtype=np.float64,
    )
    indices = np.argsort(-colors[:, 3], axis=0)
    colors = colors[indices]

    # create args
    label_metadata = np.array(
        [
            [2, 6],
            [2, 6],
            [1, 6],
            [1, 6],
        ],
        dtype=np.int32,
    )

    # compute ROCAUC and mROCAUC
    (_, _, _, _, _, rocauc, mean_rocauc) = compute_precision_recall_rocauc(
        data=colors,
        label_metadata=label_metadata,
        n_datums=6,
        score_thresholds=np.array([]),
        hardmax=False,
    )

    # test ROCAUC
    assert rocauc.shape == (label_metadata.shape[0],)
    assert rocauc[0] == 0.75  # (color, white)
    assert rocauc[1] == 0.875  # (color, red)
    assert rocauc[2] == 0.0  # (color, blue)
    assert rocauc[3] == 0.09999999999999998  # (color, black)

    # test mROCAUC
    assert mean_rocauc == 0.43125


def test_rocauc_with_animal_example(
    classifications_animal_example: list[Classification],
):

    loader = DataLoader()
    loader.add_data(classifications_animal_example)
    evaluator = loader.finalize()

    metrics = evaluator.evaluate()

    # test ROCAUC
    actual_metrics = [m.to_dict() for m in metrics[MetricType.ROCAUC]]
    expected_metrics = [
        {
            "type": "ROCAUC",
            "value": 0.7777777777777778,
            "parameters": {
                "label": "bird",
            },
        },
        {
            "type": "ROCAUC",
            "value": 0.625,
            "parameters": {
                "label": "dog",
            },
        },
        {
            "type": "ROCAUC",
            "value": 1.0,
            "parameters": {
                "label": "cat",
            },
        },
    ]
    for m in actual_metrics:
        assert m in expected_metrics
    for m in expected_metrics:
        assert m in actual_metrics

    # test mROCAUC
    actual_metrics = [m.to_dict() for m in metrics[MetricType.mROCAUC]]
    expected_metrics = [
        {"type": "mROCAUC", "value": 0.8009259259259259, "parameters": {}},
    ]
    for m in actual_metrics:
        assert m in expected_metrics
    for m in expected_metrics:
        assert m in actual_metrics


def test_rocauc_with_color_example(
    classifications_color_example: list[Classification],
):

    loader = DataLoader()
    loader.add_data(classifications_color_example)
    evaluator = loader.finalize()

    metrics = evaluator.evaluate(hardmax=False)

    # test ROCAUC
    actual_metrics = [m.to_dict() for m in metrics[MetricType.ROCAUC]]
    expected_metrics = [
        {
            "type": "ROCAUC",
            "value": 0.75,
            "parameters": {
                "label": "white",
            },
        },
        {
            "type": "ROCAUC",
            "value": 0.875,
            "parameters": {
                "label": "red",
            },
        },
        {
            "type": "ROCAUC",
            "value": 0.0,
            "parameters": {
                "label": "blue",
            },
        },
        {
            "type": "ROCAUC",
            "value": 0.09999999999999998,
            "parameters": {
                "label": "black",
            },
        },
    ]
    for m in actual_metrics:
        assert m in expected_metrics
    for m in expected_metrics:
        assert m in actual_metrics

    # test mROCAUC
    actual_metrics = [m.to_dict() for m in metrics[MetricType.mROCAUC]]
    expected_metrics = [
        {"type": "mROCAUC", "value": 0.43125, "parameters": {}},
    ]
    for m in actual_metrics:
        assert m in expected_metrics
    for m in expected_metrics:
        assert m in actual_metrics


def test_rocauc_with_image_example(
    classifications_image_example: list[Classification],
):
    loader = DataLoader()
    loader.add_data(classifications_image_example)
    loader.finalize()
    evaluator = loader.finalize()
    evaluator.evaluate()

    metrics = evaluator.evaluate()

    actual_metrics = [m.to_dict() for m in metrics[MetricType.ROCAUC]]
    expected_metrics = [
        {
            "type": "ROCAUC",
            "value": 0.0,
            "parameters": {"label": "v4"},
        },
    ]
    for m in actual_metrics:
        assert m in expected_metrics
    for m in expected_metrics:
        assert m in actual_metrics

    actual_metrics = [m.to_dict() for m in metrics[MetricType.mROCAUC]]
    expected_metrics = [
        {"type": "mROCAUC", "value": 0.0, "parameters": {}},
    ]
    for m in actual_metrics:
        assert m in expected_metrics
    for m in expected_metrics:
        assert m in actual_metrics


def test_rocauc_with_tabular_example(
    classifications_tabular_example: list[Classification],
):
    loader = DataLoader()
    loader.add_data(classifications_tabular_example)
    loader.finalize()
    evaluator = loader.finalize()
    evaluator.evaluate()

    metrics = evaluator.evaluate()

    actual_metrics = [m.to_dict() for m in metrics[MetricType.ROCAUC]]
    expected_metrics = [
        {
            "type": "ROCAUC",
            "value": 0.75,
            "parameters": {"label": "1"},
        },
        {
            "type": "ROCAUC",
            "value": 1.0,
            "parameters": {"label": "0"},
        },
        {
            "type": "ROCAUC",
            "value": 0.5555555555555556,
            "parameters": {"label": "2"},
        },
    ]
    for m in actual_metrics:
        assert m in expected_metrics
    for m in expected_metrics:
        assert m in actual_metrics

    actual_metrics = [m.to_dict() for m in metrics[MetricType.mROCAUC]]
    expected_metrics = [
        {
            "type": "mROCAUC",
            "value": 0.7685185185185185,
            "parameters": {},
        }
    ]
    for m in actual_metrics:
        assert m in expected_metrics
    for m in expected_metrics:
        assert m in actual_metrics



import pytest
from valor_lite.text_generation import Context, Evaluator, QueryResponse
from valor_lite.text_generation.computation import calculate_sentence_bleu


def test_sentence_bleu_no_context(mock_client):

    evaluator = Evaluator(client=mock_client)
    with pytest.raises(ValueError):
        assert evaluator.compute_sentence_bleu(
            response=QueryResponse(
                query="a",
                response="a",
            )
        )


def test_calculate_sentence_bleu():

    # perfect match
    assert (
        calculate_sentence_bleu(
            prediction="Mary loves Joe",
            references=["Mary loves Joe"],
            weights=(1,),
        )
        == 1.0
    )

    # perfect match, weights are a list
    assert (
        calculate_sentence_bleu(
            prediction="Mary loves Joe",
            references=["Mary loves Joe"],
            weights=[1],
        )
        == 1.0
    )

    # perfect match, case sensitive
    assert (
        calculate_sentence_bleu(
            prediction="MARY LOVES JOE",
            references=["Mary loves Joe"],
            weights=(1,),
        )
        == 0.0
    )

    # perfect match, case sensitive
    assert (
        calculate_sentence_bleu(
            prediction="Mary loves Joe",
            references=["MARY LOVES JOE"],
            weights=(1,),
        )
        == 0.0
    )

    # perfect match, case sensitive, BLEU-2
    assert (
        calculate_sentence_bleu(
            prediction="Mary loves Joe",
            references=["MARY LOVES JOE"],
            weights=(0.0, 1.0),
        )
        == 0.0
    )

    # BLEU-2
    assert (
        calculate_sentence_bleu(
            prediction="Mary loves Joe",
            references=["Mary loves Joe"],
            weights=(0, 1),
        )
        == 1.0
    )

    # BLEU-4
    assert (
        calculate_sentence_bleu(
            prediction="Mary loves Joe",
            references=["Mary loves Joe"],
            weights=[0.25] * 4,
        )
        < 1e-9
    )

    # off by one
    assert (
        calculate_sentence_bleu(
            prediction="Mary loves Joe",
            references=["Mary loves Jane"],
            weights=(1,),
        )
        == 2 / 3
    )

    # off by one BLEU-2
    assert (
        calculate_sentence_bleu(
            prediction="Mary loves Joe",
            references=["Mary loves Jane"],
            weights=(0, 1),
        )
        == 0.5
    )

    # off by one BLEU-3
    assert (
        calculate_sentence_bleu(
            prediction="Mary loves Joe",
            references=["Mary loves Jane"],
            weights=(0, 0, 1),
        )
        < 1e-9
    )

    # off by one BLEU-4
    assert (
        calculate_sentence_bleu(
            prediction="Mary loves Joe",
            references=["Mary loves Jane"],
            weights=(0, 0, 0, 1),
        )
        < 1e-9
    )

    # different cases
    assert (
        calculate_sentence_bleu(
            prediction="mary loves joe",
            references=["MARY LOVES JOE"],
            weights=(1,),
        )
        == 0.0
    )

    # different cases BLEU-2
    assert (
        calculate_sentence_bleu(
            prediction="mary loves joe",
            references=["MARY LOVES JOE"],
            weights=[0, 1],
        )
        == 0.0
    )

    # different cases BLEU-10
    assert (
        calculate_sentence_bleu(
            prediction="mary loves joe",
            references=["MARY LOVES JOE"],
            weights=[0, 0, 0, 0, 0, 0, 0, 0, 0, 1],
        )
        == 0.0
    )

    # test multiple references
    assert (
        calculate_sentence_bleu(
            prediction="flip the roaring white dolphin",
            references=[
                "some random sentence",
                "some other sentence",
                "some final reference",
                "flip the roaring white dolphin",
            ],
            weights=[0, 1],
        )
        == 1.0
    )

    # test empty weights
    with pytest.raises(ValueError):
        calculate_sentence_bleu(
            prediction="flip the roaring white dolphin",
            references=[
                "some random sentence",
            ],
            weights=[],
        )


def test_evaluate_sentence_bleu():

    evaluator = Evaluator()

    # perfect match
    assert evaluator.compute_sentence_bleu(
        response=QueryResponse(
            query="n/a",
            response="Mary loves Joe",
            context=Context(
                groundtruth=["Mary loves Joe"],
            ),
        ),
        weights=[1],
    ).to_dict() == {
        "type": "BLEU",
        "value": 1.0,
        "parameters": {
            "weights": [1],
        },
    }

    # off by one
    assert evaluator.compute_sentence_bleu(
        response=QueryResponse(
            query="n/a",
            response="Mary loves Joe",
            context=Context(
                groundtruth=["Mary loves Jane"],
            ),
        ),
        weights=[1],
    ).to_dict() == {
        "type": "BLEU",
        "value": 2 / 3,
        "parameters": {
            "weights": [1],
        },
    }



from .annotation import Context, QueryResponse
from .llm.integrations import ClientWrapper, MistralWrapper, OpenAIWrapper
from .manager import Evaluator
from .metric import Metric, MetricType

__all__ = [
    "QueryResponse",
    "Context",
    "Evaluator",
    "Metric",
    "MetricType",
    "ClientWrapper",
    "OpenAIWrapper",
    "MistralWrapper",
]



from dataclasses import replace

import numpy as np
import pytest
from valor_lite.object_detection import (
    BoundingBox,
    DataLoader,
    Detection,
    MetricType,
)


@pytest.fixture
def one_detection(basic_detections: list[Detection]) -> list[Detection]:
    return [basic_detections[0]]


@pytest.fixture
def two_detections(basic_detections: list[Detection]) -> list[Detection]:
    return basic_detections


@pytest.fixture
def four_detections(basic_detections: list[Detection]) -> list[Detection]:
    det1 = basic_detections[0]
    det2 = basic_detections[1]
    det3 = replace(basic_detections[0])
    det4 = replace(basic_detections[1])

    det3.uid = "uid3"
    det4.uid = "uid4"

    return [det1, det2, det3, det4]


def _generate_random_detections(
    n_detections: int, n_boxes: int, labels: str
) -> list[Detection]:
    from random import choice, uniform

    def bbox(is_prediction):
        xmin, ymin = uniform(0, 10), uniform(0, 10)
        xmax, ymax = uniform(xmin, 15), uniform(ymin, 15)
        kw = {"scores": [uniform(0, 1)]} if is_prediction else {}
        return BoundingBox(
            xmin,
            xmax,
            ymin,
            ymax,
            [choice(labels)],
            **kw,
        )

    return [
        Detection(
            uid=f"uid{i}",
            groundtruths=[bbox(is_prediction=False) for _ in range(n_boxes)],
            predictions=[bbox(is_prediction=True) for _ in range(n_boxes)],
        )
        for i in range(n_detections)
    ]


def test_filtering_one_detection(one_detection: list[Detection]):
    """
    Basic object detection test that combines the labels of basic_detections_first_class and basic_detections_second_class.

    groundtruths
        datum uid1
            box 1 - label v1 - tp
            box 3 - label v2 - fn missing prediction

    predictions
        datum uid1
            box 1 - label v1 - score 0.3 - tp
    """

    loader = DataLoader()
    loader.add_bounding_boxes(one_detection)
    evaluator = loader.finalize()

    assert (
        evaluator._ranked_pairs
        == np.array(
            [
                [0.0, 0.0, 0.0, 1.0, 0.0, 0.0, 0.3],
            ]
        )
    ).all()

    assert (
        evaluator._label_metadata_per_datum
        == np.array(
            [
                [
                    [1, 1],
                ],
                [
                    [1, 0],
                ],
            ]
        )
    ).all()

    assert (evaluator._label_metadata == np.array([[1, 1], [1, 0]])).all()

    # test datum filtering
    filter_ = evaluator.create_filter(datum_uids=["uid1"])
    assert (filter_.ranked_indices == np.array([0])).all()
    assert (filter_.detailed_indices == np.array([0, 1])).all()
    assert (
        filter_.label_metadata
        == np.array(
            [
                [
                    1,
                    1,
                ],
                [1, 0],
            ]
        )
    ).all()

    with pytest.raises(KeyError) as e:
        evaluator.create_filter(datum_uids=["uid2"])
    assert "uid2" in str(e)

    # test label filtering
    filter_ = evaluator.create_filter(labels=["v1"])
    assert (filter_.ranked_indices == np.array([0])).all()
    assert (filter_.detailed_indices == np.array([0])).all()
    assert (filter_.label_metadata == np.array([[1, 1], [0, 0]])).all()

    filter_ = evaluator.create_filter(labels=["v2"])
    assert (filter_.ranked_indices == np.array([])).all()
    assert (filter_.detailed_indices == np.array([1])).all()
    assert (filter_.label_metadata == np.array([[0, 0], [1, 0]])).all()

    # test combo
    filter_ = evaluator.create_filter(
        datum_uids=["uid1"],
        labels=["v1"],
    )
    assert (filter_.ranked_indices == np.array([0])).all()
    assert (filter_.detailed_indices == np.array([0])).all()
    assert (filter_.label_metadata == np.array([[1, 1], [0, 0]])).all()

    # test evaluation
    filter_ = evaluator.create_filter(datum_uids=["uid1"])
    metrics = evaluator.evaluate(
        iou_thresholds=[0.5],
        filter_=filter_,
    )

    actual_metrics = [m.to_dict() for m in metrics[MetricType.AP]]
    expected_metrics = [
        {
            "type": "AP",
            "value": 1.0,
            "parameters": {
                "iou_threshold": 0.5,
                "label": "v1",
            },
        },
        {
            "type": "AP",
            "value": 0.0,
            "parameters": {
                "iou_threshold": 0.5,
                "label": "v2",
            },
        },
    ]
    for m in actual_metrics:
        assert m in expected_metrics
    for m in expected_metrics:
        assert m in actual_metrics


def test_filtering_two_detections(two_detections: list[Detection]):
    """
    Basic object detection test that combines the labels of basic_detections_first_class and basic_detections_second_class.

    groundtruths
        datum uid1
            box 1 - label v1 - tp
            box 3 - label v2 - fn missing prediction
        datum uid2
            box 2 - label v1 - fn misclassification

    predictions
        datum uid1
            box 1 - label v1 - score 0.3 - tp
        datum uid2
            box 2 - label v2 - score 0.98 - fp misclassification
    """

    loader = DataLoader()
    loader.add_bounding_boxes(two_detections)
    evaluator = loader.finalize()

    assert (
        evaluator._ranked_pairs
        == np.array(
            [
                [1.0, 0.0, 0.0, 1.0, 0.0, 1.0, 0.98],
                [0.0, 0.0, 0.0, 1.0, 0.0, 0.0, 0.3],
            ]
        )
    ).all()

    assert (
        evaluator._label_metadata_per_datum
        == np.array(
            [
                [
                    [1, 1],
                    [1, 0],
                ],
                [
                    [1, 0],
                    [0, 1],
                ],
            ]
        )
    ).all()

    assert (evaluator._label_metadata == np.array([[2, 1], [1, 1]])).all()

    # test datum filtering
    filter_ = evaluator.create_filter(datum_uids=["uid1"])
    assert (filter_.ranked_indices == np.array([1])).all()
    assert (filter_.detailed_indices == np.array([0, 1])).all()
    assert (filter_.label_metadata == np.array([[1, 1], [1, 0]])).all()

    filter_ = evaluator.create_filter(datum_uids=["uid2"])
    assert (filter_.ranked_indices == np.array([0])).all()
    assert (filter_.detailed_indices == np.array([2])).all()
    assert (
        filter_.label_metadata
        == np.array(
            [
                [
                    1,
                    0,
                ],
                [
                    0,
                    1,
                ],
            ]
        )
    ).all()

    # test label filtering
    filter_ = evaluator.create_filter(labels=["v1"])
    assert (filter_.ranked_indices == np.array([0, 1])).all()
    assert (filter_.detailed_indices == np.array([0, 2])).all()
    assert (filter_.label_metadata == np.array([[2, 1], [0, 0]])).all()

    filter_ = evaluator.create_filter(labels=["v2"])
    assert (filter_.ranked_indices == np.array([])).all()
    assert (filter_.detailed_indices == np.array([1])).all()
    assert (filter_.label_metadata == np.array([[0, 0], [1, 1]])).all()

    # test combo
    filter_ = evaluator.create_filter(
        datum_uids=["uid1"],
        labels=["v1"],
    )
    assert (filter_.ranked_indices == np.array([1])).all()
    assert (filter_.detailed_indices == np.array([0])).all()
    assert (filter_.label_metadata == np.array([[1, 1], [0, 0]])).all()

    # test evaluation
    filter_ = evaluator.create_filter(datum_uids=["uid1"])
    metrics = evaluator.evaluate(
        iou_thresholds=[0.5],
        filter_=filter_,
    )

    actual_metrics = [m.to_dict() for m in metrics[MetricType.AP]]
    expected_metrics = [
        {
            "type": "AP",
            "value": 1.0,
            "parameters": {"iou_threshold": 0.5, "label": "v1"},
        },
        {
            "type": "AP",
            "value": 0.0,
            "parameters": {"iou_threshold": 0.5, "label": "v2"},
        },
    ]
    for m in actual_metrics:
        assert m in expected_metrics
    for m in expected_metrics:
        assert m in actual_metrics


def test_filtering_four_detections(four_detections: list[Detection]):
    """
    Basic object detection test that combines the labels of basic_detections_first_class and basic_detections_second_class.

    groundtruths
        datum uid1
            box 1 - label v1 - tp
            box 3 - label v2 - fn missing prediction
        datum uid2
            box 2 - label v1 - fn misclassification
        datum uid3
            box 1 - label v1 - tp
            box 3 - label v2 - fn missing prediction
        datum uid4
            box 2 - label v1 - fn misclassification

    predictions
        datum uid1
            box 1 - label v1 - score 0.3 - tp
        datum uid2
            box 2 - label v2 - score 0.98 - fp misclassification
        datum uid3
            box 1 - label v1 - score 0.3 - tp
        datum uid4
            box 2 - label v2 - score 0.98 - fp misclassification
    """

    loader = DataLoader()
    loader.add_bounding_boxes(four_detections)
    evaluator = loader.finalize()

    assert (
        evaluator._ranked_pairs
        == np.array(
            [
                [1.0, 0.0, 0.0, 1.0, 0.0, 1.0, 0.98],
                [3.0, 0.0, 0.0, 1.0, 0.0, 1.0, 0.98],
                [0.0, 0.0, 0.0, 1.0, 0.0, 0.0, 0.3],
                [2.0, 0.0, 0.0, 1.0, 0.0, 0.0, 0.3],
            ]
        )
    ).all()

    assert (
        evaluator._label_metadata_per_datum
        == np.array(
            [
                [
                    [1, 1],
                    [1, 0],
                    [1, 1],
                    [1, 0],
                ],
                [
                    [1, 0],
                    [0, 1],
                    [1, 0],
                    [0, 1],
                ],
            ],
            dtype=np.int32,
        )
    ).all()

    assert (evaluator._label_metadata == np.array([[4, 2], [2, 2]])).all()

    # test datum filtering
    filter_ = evaluator.create_filter(datum_uids=["uid1"])
    assert (filter_.ranked_indices == np.array([2])).all()
    assert (filter_.detailed_indices == np.array([0, 1])).all()
    assert (filter_.label_metadata == np.array([[1, 1], [1, 0]])).all()

    filter_ = evaluator.create_filter(datum_uids=["uid2"])
    assert (filter_.ranked_indices == np.array([0])).all()
    assert (filter_.detailed_indices == np.array([2])).all()
    assert (filter_.label_metadata == np.array([[1, 0], [0, 1]])).all()

    # test label filtering
    filter_ = evaluator.create_filter(labels=["v1"])
    assert (filter_.ranked_indices == np.array([0, 1, 2, 3])).all()
    assert (filter_.detailed_indices == np.array([0, 2, 3, 5])).all()
    assert (filter_.label_metadata == np.array([[4, 2], [0, 0]])).all()

    filter_ = evaluator.create_filter(labels=["v2"])
    assert (filter_.ranked_indices == np.array([])).all()
    assert (filter_.detailed_indices == np.array([1, 4])).all()
    assert (filter_.label_metadata == np.array([[0, 0], [2, 2]])).all()

    # test combo
    filter_ = evaluator.create_filter(
        datum_uids=["uid1"],
        labels=["v1"],
    )
    assert (filter_.ranked_indices == np.array([2])).all()
    assert (filter_.detailed_indices == np.array([0])).all()
    assert (
        filter_.label_metadata
        == np.array(
            [
                [
                    1,
                    1,
                ],
                [
                    0,
                    0,
                ],
            ]
        )
    ).all()

    # test evaluation
    filter_ = evaluator.create_filter(datum_uids=["uid1"])

    metrics = evaluator.evaluate(
        iou_thresholds=[0.5],
        filter_=filter_,
    )

    actual_metrics = [m.to_dict() for m in metrics[MetricType.AP]]
    expected_metrics = [
        {
            "type": "AP",
            "value": 1.0,
            "parameters": {"iou_threshold": 0.5, "label": "v1"},
        },
        {
            "type": "AP",
            "value": 0.0,
            "parameters": {"iou_threshold": 0.5, "label": "v2"},
        },
    ]
    for m in actual_metrics:
        assert m in expected_metrics
    for m in expected_metrics:
        assert m in actual_metrics


def test_filtering_all_detections(four_detections: list[Detection]):
    """
    Basic object detection test that combines the labels of basic_detections_first_class and basic_detections_second_class.

    groundtruths
        datum uid1
            box 1 - label v1 - tp
            box 3 - label v2 - fn missing prediction
        datum uid2
            box 2 - label v1 - fn misclassification
        datum uid3
            box 1 - label v1 - tp
            box 3 - label v2 - fn missing prediction
        datum uid4
            box 2 - label v1 - fn misclassification

    predictions
        datum uid1
            box 1 - label v1 - score 0.3 - tp
        datum uid2
            box 2 - label v2 - score 0.98 - fp misclassification
        datum uid3
            box 1 - label v1 - score 0.3 - tp
        datum uid4
            box 2 - label v2 - score 0.98 - fp misclassification
    """

    loader = DataLoader()
    loader.add_bounding_boxes(four_detections)
    evaluator = loader.finalize()

    assert (
        evaluator._ranked_pairs
        == np.array(
            [
                [1.0, 0.0, 0.0, 1.0, 0.0, 1.0, 0.98],
                [3.0, 0.0, 0.0, 1.0, 0.0, 1.0, 0.98],
                [0.0, 0.0, 0.0, 1.0, 0.0, 0.0, 0.3],
                [2.0, 0.0, 0.0, 1.0, 0.0, 0.0, 0.3],
            ]
        )
    ).all()

    assert (
        evaluator._label_metadata_per_datum
        == np.array(
            [
                [
                    [1, 1],
                    [1, 0],
                    [1, 1],
                    [1, 0],
                ],
                [
                    [1, 0],
                    [0, 1],
                    [1, 0],
                    [0, 1],
                ],
            ],
            dtype=np.int32,
        )
    ).all()

    assert (evaluator._label_metadata == np.array([[4, 2], [2, 2]])).all()

    # test datum filtering
    filter_ = evaluator.create_filter(datum_uids=[])
    assert (filter_.ranked_indices == np.array([])).all()
    assert (filter_.detailed_indices == np.array([])).all()
    assert (
        filter_.label_metadata
        == np.array(
            [
                [
                    0,
                    0,
                ],
                [
                    0,
                    0,
                ],
            ]
        )
    ).all()

    # test label filtering
    filter_ = evaluator.create_filter(labels=[])
    assert (filter_.ranked_indices == np.array([])).all()
    assert (filter_.detailed_indices == np.array([])).all()
    assert (
        filter_.label_metadata
        == np.array(
            [
                [
                    0,
                    0,
                ],
                [
                    0,
                    0,
                ],
            ]
        )
    ).all()

    # test combo
    filter_ = evaluator.create_filter(
        datum_uids=[],
        labels=["v1"],
    )
    assert (filter_.ranked_indices == np.array([])).all()
    assert (filter_.detailed_indices == np.array([])).all()
    assert (
        filter_.label_metadata
        == np.array(
            [
                [
                    0,
                    0,
                ],
                [0, 0],
            ]
        )
    ).all()

    # test evaluation
    filter_ = evaluator.create_filter(datum_uids=[])

    metrics = evaluator.evaluate(
        iou_thresholds=[0.5],
        filter_=filter_,
    )
    evaluator.compute_confusion_matrix(
        iou_thresholds=[0.5],
        filter_=filter_,
    )

    actual_metrics = [m.to_dict() for m in metrics[MetricType.AP]]
    assert len(actual_metrics) == 0


def test_filtering_random_detections():
    loader = DataLoader()
    loader.add_bounding_boxes(_generate_random_detections(13, 4, "abc"))
    evaluator = loader.finalize()
    f = evaluator.create_filter(datum_uids=["uid1"])
    evaluator.evaluate(filter_=f)



from collections import defaultdict

import numpy as np
from numpy.typing import NDArray
from valor_lite.classification.metric import Metric, MetricType


def unpack_precision_recall_rocauc_into_metric_lists(
    results: tuple[
        NDArray[np.int32],
        NDArray[np.float64],
        NDArray[np.float64],
        NDArray[np.float64],
        NDArray[np.float64],
        NDArray[np.float64],
        float,
    ],
    score_thresholds: list[float],
    hardmax: bool,
    label_metadata: NDArray[np.int32],
    index_to_label: dict[int, str],
) -> dict[MetricType, list[Metric]]:
    (
        counts,
        precision,
        recall,
        accuracy,
        f1_score,
        rocauc,
        mean_rocauc,
    ) = results

    metrics = defaultdict(list)

    metrics[MetricType.ROCAUC] = [
        Metric.roc_auc(
            value=float(rocauc[label_idx]),
            label=label,
        )
        for label_idx, label in index_to_label.items()
        if label_metadata[label_idx, 0] > 0
    ]

    metrics[MetricType.mROCAUC] = [
        Metric.mean_roc_auc(
            value=float(mean_rocauc),
        )
    ]

    metrics[MetricType.Accuracy] = [
        Metric.accuracy(
            value=float(accuracy[score_idx]),
            score_threshold=score_threshold,
            hardmax=hardmax,
        )
        for score_idx, score_threshold in enumerate(score_thresholds)
    ]

    for label_idx, label in index_to_label.items():
        for score_idx, score_threshold in enumerate(score_thresholds):

            kwargs = {
                "label": label,
                "hardmax": hardmax,
                "score_threshold": score_threshold,
            }
            row = counts[:, label_idx]
            metrics[MetricType.Counts].append(
                Metric.counts(
                    tp=int(row[score_idx, 0]),
                    fp=int(row[score_idx, 1]),
                    fn=int(row[score_idx, 2]),
                    tn=int(row[score_idx, 3]),
                    **kwargs,
                )
            )

            # if no groundtruths exists for a label, skip it.
            if label_metadata[label_idx, 0] == 0:
                continue

            metrics[MetricType.Precision].append(
                Metric.precision(
                    value=float(precision[score_idx, label_idx]),
                    **kwargs,
                )
            )
            metrics[MetricType.Recall].append(
                Metric.recall(
                    value=float(recall[score_idx, label_idx]),
                    **kwargs,
                )
            )
            metrics[MetricType.F1].append(
                Metric.f1_score(
                    value=float(f1_score[score_idx, label_idx]),
                    **kwargs,
                )
            )
    return metrics


def _unpack_confusion_matrix_value(
    confusion_matrix: NDArray[np.float64],
    number_of_labels: int,
    number_of_examples: int,
    index_to_uid: dict[int, str],
    index_to_label: dict[int, str],
) -> dict[str, dict[str, dict[str, int | list[dict[str, str | float]]]]]:
    """
    Unpacks a numpy array of confusion matrix counts and examples.
    """

    datum_idx = lambda gt_label_idx, pd_label_idx, example_idx: int(  # noqa: E731 - lambda fn
        confusion_matrix[
            gt_label_idx,
            pd_label_idx,
            example_idx * 2 + 1,
        ]
    )

    score_idx = lambda gt_label_idx, pd_label_idx, example_idx: float(  # noqa: E731 - lambda fn
        confusion_matrix[
            gt_label_idx,
            pd_label_idx,
            example_idx * 2 + 2,
        ]
    )

    return {
        index_to_label[gt_label_idx]: {
            index_to_label[pd_label_idx]: {
                "count": max(
                    int(confusion_matrix[gt_label_idx, pd_label_idx, 0]),
                    0,
                ),
                "examples": [
                    {
                        "datum": index_to_uid[
                            datum_idx(gt_label_idx, pd_label_idx, example_idx)
                        ],
                        "score": score_idx(
                            gt_label_idx, pd_label_idx, example_idx
                        ),
                    }
                    for example_idx in range(number_of_examples)
                    if datum_idx(gt_label_idx, pd_label_idx, example_idx) >= 0
                ],
            }
            for pd_label_idx in range(number_of_labels)
        }
        for gt_label_idx in range(number_of_labels)
    }


def _unpack_missing_predictions_value(
    missing_predictions: NDArray[np.int32],
    number_of_labels: int,
    number_of_examples: int,
    index_to_uid: dict[int, str],
    index_to_label: dict[int, str],
) -> dict[str, dict[str, int | list[dict[str, str]]]]:
    """
    Unpacks a numpy array of missing prediction counts and examples.
    """

    datum_idx = (
        lambda gt_label_idx, example_idx: int(  # noqa: E731 - lambda fn
            missing_predictions[
                gt_label_idx,
                example_idx + 1,
            ]
        )
    )

    return {
        index_to_label[gt_label_idx]: {
            "count": max(
                int(missing_predictions[gt_label_idx, 0]),
                0,
            ),
            "examples": [
                {"datum": index_to_uid[datum_idx(gt_label_idx, example_idx)]}
                for example_idx in range(number_of_examples)
                if datum_idx(gt_label_idx, example_idx) >= 0
            ],
        }
        for gt_label_idx in range(number_of_labels)
    }


def unpack_confusion_matrix_into_metric_list(
    results: tuple[NDArray[np.float64], NDArray[np.int32]],
    score_thresholds: list[float],
    number_of_examples: int,
    index_to_uid: dict[int, str],
    index_to_label: dict[int, str],
) -> list[Metric]:

    (confusion_matrix, missing_predictions) = results
    n_scores, n_labels, _, _ = confusion_matrix.shape
    return [
        Metric.confusion_matrix(
            score_threshold=score_threshold,
            maximum_number_of_examples=number_of_examples,
            confusion_matrix=_unpack_confusion_matrix_value(
                confusion_matrix=confusion_matrix[score_idx, :, :, :],
                number_of_labels=n_labels,
                number_of_examples=number_of_examples,
                index_to_label=index_to_label,
                index_to_uid=index_to_uid,
            ),
            missing_predictions=_unpack_missing_predictions_value(
                missing_predictions=missing_predictions[score_idx, :, :],
                number_of_labels=n_labels,
                number_of_examples=number_of_examples,
                index_to_label=index_to_label,
                index_to_uid=index_to_uid,
            ),
        )
        for score_idx, score_threshold in enumerate(score_thresholds)
    ]



import numpy as np
import pytest
from valor_lite.semantic_segmentation import (
    Bitmask,
    Segmentation,
    generate_segmentation,
)


def test_bitmask():

    Bitmask(
        mask=np.array([True, False]),
        label="label",
    )

    with pytest.raises(ValueError) as e:
        Bitmask(mask=np.array([1, 2]), label="label")
    assert "int64" in str(e)


def test_segmentation():

    s = Segmentation(
        uid="uid",
        groundtruths=[
            Bitmask(
                mask=np.array([True, False]),
                label="label",
            )
        ],
        predictions=[
            Bitmask(
                mask=np.array([True, False]),
                label="label",
            )
        ],
    )
    assert s.shape == (2,)
    assert s.size == 2

    with pytest.raises(ValueError) as e:
        Segmentation(
            uid="uid",
            groundtruths=[
                Bitmask(
                    mask=np.array([True, False, False]),
                    label="label",
                )
            ],
            predictions=[
                Bitmask(
                    mask=np.array([True, False]),
                    label="label",
                )
            ],
        )
    assert "mismatch" in str(e)

    with pytest.raises(ValueError) as e:
        Segmentation(
            uid="uid",
            groundtruths=[],
            predictions=[
                Bitmask(
                    mask=np.array([True, False]),
                    label="label",
                )
            ],
        )
    assert "missing ground truths" in str(e)

    with pytest.raises(ValueError) as e:
        Segmentation(
            uid="uid",
            groundtruths=[
                Bitmask(
                    mask=np.array([True, False]),
                    label="label",
                )
            ],
            predictions=[],
        )
    assert "missing predictions" in str(e)


def test_generate_segmentation():

    # N labels > 1
    segmentation = generate_segmentation(
        datum_uid="uid1",
        number_of_unique_labels=3,
        mask_height=2,
        mask_width=3,
    )

    assert segmentation.uid == "uid1"
    assert segmentation.shape == (2, 3)
    assert segmentation.size == 6

    assert len(segmentation.groundtruths) == 3
    assert all(gt.mask.dtype == np.bool_ for gt in segmentation.groundtruths)
    assert all(gt.mask.shape == (2, 3) for gt in segmentation.groundtruths)

    assert len(segmentation.predictions) == 3
    assert all(pd.mask.dtype == np.bool_ for pd in segmentation.predictions)
    assert all(pd.mask.shape == (2, 3) for pd in segmentation.predictions)

    # N labels = 1
    segmentation = generate_segmentation(
        datum_uid="uid1",
        number_of_unique_labels=1,
        mask_height=2,
        mask_width=3,
    )

    assert segmentation.uid == "uid1"
    assert segmentation.shape == (2, 3)
    assert segmentation.size == 6

    assert len(segmentation.groundtruths) == 1
    assert all(gt.mask.dtype == np.bool_ for gt in segmentation.groundtruths)
    assert all(gt.mask.shape == (2, 3) for gt in segmentation.groundtruths)

    assert len(segmentation.predictions) == 1
    assert all(pd.mask.dtype == np.bool_ for pd in segmentation.predictions)
    assert all(pd.mask.shape == (2, 3) for pd in segmentation.predictions)

    # N labels = 0
    with pytest.raises(ValueError):
        generate_segmentation(
            datum_uid="uid1",
            number_of_unique_labels=0,
            mask_height=2,
            mask_width=3,
        )



from .annotation import Bitmask, Segmentation, generate_segmentation
from .manager import DataLoader, Evaluator
from .metric import Metric, MetricType

__all__ = [
    "DataLoader",
    "Evaluator",
    "Segmentation",
    "Bitmask",
    "Metric",
    "MetricType",
    "generate_segmentation",
]



from collections import defaultdict

import numpy as np
from numpy.typing import NDArray
from valor_lite.semantic_segmentation.metric import Metric, MetricType


def unpack_precision_recall_iou_into_metric_lists(
    results: tuple,
    label_metadata: NDArray[np.int32],
    index_to_label: dict[int, str],
) -> dict[MetricType, list[Metric]]:

    n_labels = len(index_to_label)
    (
        precision,
        recall,
        f1_score,
        accuracy,
        ious,
        hallucination_ratios,
        missing_prediction_ratios,
    ) = results

    metrics = defaultdict(list)

    metrics[MetricType.Accuracy] = [
        Metric.accuracy(
            value=float(accuracy),
        )
    ]

    metrics[MetricType.ConfusionMatrix] = [
        Metric.confusion_matrix(
            confusion_matrix={
                index_to_label[gt_label_idx]: {
                    index_to_label[pd_label_idx]: {
                        "iou": float(ious[gt_label_idx, pd_label_idx])
                    }
                    for pd_label_idx in range(n_labels)
                    if label_metadata[pd_label_idx, 0] > 0
                }
                for gt_label_idx in range(n_labels)
                if label_metadata[gt_label_idx, 0] > 0
            },
            hallucinations={
                index_to_label[pd_label_idx]: {
                    "ratio": float(hallucination_ratios[pd_label_idx])
                }
                for pd_label_idx in range(n_labels)
                if label_metadata[pd_label_idx, 0] > 0
            },
            missing_predictions={
                index_to_label[gt_label_idx]: {
                    "ratio": float(missing_prediction_ratios[gt_label_idx])
                }
                for gt_label_idx in range(n_labels)
                if label_metadata[gt_label_idx, 0] > 0
            },
        )
    ]

    metrics[MetricType.mIOU] = [
        Metric.mean_iou(
            value=float(ious.diagonal().mean()),
        )
    ]

    for label_idx, label in index_to_label.items():

        kwargs = {
            "label": label,
        }

        # if no groundtruths exists for a label, skip it.
        if label_metadata[label_idx, 0] == 0:
            continue

        metrics[MetricType.Precision].append(
            Metric.precision(
                value=float(precision[label_idx]),
                **kwargs,
            )
        )
        metrics[MetricType.Recall].append(
            Metric.recall(
                value=float(recall[label_idx]),
                **kwargs,
            )
        )
        metrics[MetricType.F1].append(
            Metric.f1_score(
                value=float(f1_score[label_idx]),
                **kwargs,
            )
        )
        metrics[MetricType.IOU].append(
            Metric.iou(
                value=float(ious[label_idx, label_idx]),
                **kwargs,
            )
        )

    return metrics
